// $Id: BasePair.h 628 2011-08-31 15:42:16Z ge $
/// \file BasePair.h
/// \brief contains the BasePair class
///
/// $Revision: 628 $
/// \author Gerald Weber

#ifndef GBC_EXP_BASEPAIR_H
#define GBC_EXP_BASEPAIR_H "$Id: BasePair.h 628 2011-08-31 15:42:16Z ge $"

#include "JobControl.h"
#include "Nucleotide.h"

namespace gbc {

template<class _InternalTp=char>
/// \brief Forms a base pair out of two Nucleotides.
/// 
/// This class forms a std::pair from Nucleotide<_InternalTp>.
class BasePair: public std::pair<Nucleotide<_InternalTp>,Nucleotide<_InternalTp> >
  {
  public:
    typedef _InternalTp internal_type;                ///< type of _InternalTp
    typedef Nucleotide<_InternalTp> nucleotide_type;  ///< type of Nucleotide<_InternalTp>
    typedef std::pair<Nucleotide<_InternalTp>,Nucleotide<_InternalTp> > pair_type;
    typedef std::pair<internal_type,internal_type> internal_pair_type;

    static const bool not_context_dependent =false;
    static const bool context_dependent     =true;

    static const bool default_Context_Dependence=not_context_dependent;

    bool Context_Dependence;        ///< Base pairs may be context dependendent 
    std::string separation_string;  ///< Separator string for the string representation.

  BasePair(void): Context_Dependence(not_context_dependent) {separation_string="";}

  BasePair(char fst, char scnd, bool context=not_context_dependent)
  /// \brief Concstructs a BasePair from its named Nucleotides.
  ///
  /// This constructor works by converting the char representation
  /// to a type Nucelotide which is then assigned to the BasePair.
   : pair_type(nucleotide_type(fst),nucleotide_type(scnd)),
     Context_Dependence(context), separation_string()
   {}

  BasePair(char fst, bool context=not_context_dependent)
  /// \brief Concstructs a BasePair from its named Nucleotide.
  ///
  /// This constructor works by converting the char representation
  /// to a type Nucelotide which is then assigned to the BasePair.
   : pair_type(nucleotide_type(fst),nucleotide_type(fst)),
     Context_Dependence(context), separation_string()
   {
   this->second.complementary();
   }

  BasePair(std::string bpname, bool context=default_Context_Dependence)
  /// \brief Concstructs a BasePair from its named Nucleotides.
  ///
  /// This constructor works by converting the string representation
  /// to a type Nucelotide which is then assigned to the BasePair.
   : pair_type(nucleotide_type(bpname[0]),nucleotide_type(bpname[1])),
     Context_Dependence(context), separation_string()
   {}

  template<class _InternalTp2>
  /// \brief Constructs a BasePair from Nucleotide and its complement.
  BasePair(const Nucleotide<_InternalTp2> fst, bool context=default_Context_Dependence)
    : pair_type(fst,fst),
      Context_Dependence(context), separation_string()
    {
    this->second.complementary();
    }

  template<class _InternalTp2>
  /// \brief Assigns two Nucleotide objects to form a BasePair.
  BasePair(const Nucleotide<_InternalTp2> fst, const Nucleotide<_InternalTp2> scnd, bool context=default_Context_Dependence)
    : pair_type(fst,scnd), 
      Context_Dependence(context), separation_string()
    {}

  template<class _InternalTp2>
  /// \brief Assigns a nucleotide, and its complement, to the BasePair.
  BasePair& operator=(const Nucleotide<_InternalTp2>& nuc)
    {
    this->first=nuc;
    this->second=nuc;
    this->second.complementary();

    return *this;
    }

  /// \brief Swaps the two Nucleotides of the Base Pairs.
  void swap(void)
    {
    nucleotide_type tmp(this->second);
    this->second=this->first;
    this->first=tmp;
    }

  /// \brief Converts this BasePair to it's cheapest symmetry equivalent.
  ///
  /// We apply the symmetry operations of swapping.
  /// Then we retain the one that yields the lowest internal
  /// representation value using the operator<. 
  ///
  /// \attention No operation is carried out if Context_Dependence is true
  inline bool reduce_to_smallest_symmetry(void)
    {
    bool reduced=false;
    if (Context_Dependence==not_context_dependent)
      { 
      BasePair sw=*this;
      sw.swap();
      if (sw < *this) {*this=sw; reduced=true;}
      }
    return reduced;
    }

  /// \brief Produces a string which describes the BasePair.
  ///
  /// This method makes use of the separation_string which by default is set to a 
  /// empty string. Possible uses are "/" which would give results such as
  /// "A/T" or "C/G". 
  operator std::string(void) const
    {
    std::string res=this->first.symbol()+separation_string+this->second.symbol();
    return res;
    }

  operator internal_pair_type(void) const
    {return internal_pair_type(this->first.representation,this->second.representation);}
  
  /// \brief Swaps the two Nucleotides of the Base Pairs.
  ///
  /// This method is called axis_swap such that it is later not
  /// confused with stack_swap of the BasePairNeighbours class.
  friend inline BasePair axis_swap(const BasePair& bp)
    {
    return BasePair(bp.second,bp.first);
    }

  /// \brief Checks if two base pairs are axis-symmetric to each other. 
  /// \attention This is done regardless of the 3'->5' direction by
  /// simply checking the the representations
  friend inline bool axis_symmetry(const BasePair& bp1, const BasePair& bp2)
    {
    return   (bp1.first.representation  == bp2.second.representation)
          && (bp1.second.representation == bp2.first.representation);
    }

  /// \brief Cheks if the two nucleotides are the same.
  friend inline bool check_symmetry(const BasePair& bp)
    {
    return bp.first.representation == bp.second.representation;
    }

  friend inline BasePair representation(internal_pair_type pr)
    {
    Nucleotide<internal_type> N1, N2;
    N1.representation=pr.first; N2.representation=pr.second;
    return BasePair(N1,N2);
    }

  /// Extractor for printing BasePair symbols
  ///
  /// This method first converts the BasePair to its string representation.
  friend std::ostream& operator<<(std::ostream &out,const BasePair &bp)
    { 
    out << static_cast<std::string>(bp);
    return out;
    }
  };//ends class


};//ends namespace
#endif
