// $Id: HeterogenousTM.h 702 2013-02-06 16:19:28Z ge $
/// \file HeterogenousTM.h
/// \brief contains the class for Heterogenous Transfer Matrix calculations
///
/// $Revision: 702 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_EXP_HETEROGENOUSTM_H
#define GBC_EXP_HETEROGENOUSTM_H "$Id: HeterogenousTM.h 702 2013-02-06 16:19:28Z ge $"
#include "TransferMatrix.h"
#include "Duplex.h"
#include "NeighbourSequence.h"
#include "ReferenceSet.h"
#include <gsl/gsl_math.h>
#include <boost/filesystem.hpp>
#include "Minimization.h"
#undef max
#undef min

namespace gbc_exp
  {
  template<class _Tp=double>
  class HeterogenousTM
    {
    public:
      typedef _Tp                                        value_type;
      typedef TransferMatrix<value_type>                 transfer_matrix_type;
      typedef ParameterMap<value_type>                   parameter_map_type;
      typedef std::string                                bp_id_type;
      typedef std::map<bp_id_type,transfer_matrix_type> tm_map_type;
      typedef std::valarray<value_type>                  vector_type;
      typedef valmatrix<value_type>                      matrix_type;
      typedef std::map<bp_id_type,matrix_type>           matrix_map_type;
      typedef std::map<bp_id_type,Range<double> >        range_map_type;
      typedef ReferenceSet<Duplex<> >                    duplex_ref_type;
      typedef ReferenceSet<NeighbourSequence<> >         neighbours_ref_type;
      typedef std::deque<std::string>                    string_deque_type;
      typedef typename transfer_matrix_type::quadrature_type      quadrature_type;

    protected:
      value_type Temperature;               ///< The temperature to be used for the hamiltonian

    public:
    bool             Periodic_boundary_condition; ///< Controls the boundary condition
    bool             Debug;
    bool             Verbose;
 
    tm_map_type      TM_map;                      ///< Map of TransferMatrices
    range_map_type   Range_map;

    std::string      ParameterIdName;
    std::string      MatrixDir;

    matrix_map_type  C_matrix_map;                ///< Map of \f$ C \f$ matrices
    matrix_map_type  Delta_matrix_map;            ///< Map of \f$ \Delta \f$ matrices
    matrix_map_type  ADelta_matrix_map;           ///< Map of \f$ \Delta \f$ matrices
    matrix_map_type* Loop_matrix_map;
    matrix_type      Y;                           ///< The \f$ Y \f$ matrix
    matrix_type      Eigenvector;
    matrix_type      Eigenvalue;
    matrix_type      Lambda_matrix;               ///< The \f$ \Lambda \f$ matrix

    typename tm_map_type::iterator Base;          ///< Iterator which points to the expansion base
    bp_id_type                     BaseName;      ///< Name of the base pair neighbours on which the expansion take place

    parameter_map_type parameter_map;



    Duplex<>             Equivalent_duplex;       ///< The translated Duplex
    NeighbourSequence<>  Equivalent_neighbours;   ///< The translated sequence of base pair neighbours
    duplex_ref_type      Duplex_set;              ///< The set of different base pairs
    neighbours_ref_type  Neighbours_set;          ///< The set of different base pair neighbours
    quadrature_type*     Quadrature;              ///< The quadrature derived from TransferMatrix

    HeterogenousTM(void): Temperature(), Periodic_boundary_condition(false), Debug(false)
      {
      Quadrature = new quadrature_type();
      Quadrature->Rule.Save_calculations=true;
      Quadrature->Rule.Load_calculations=true;
      Quadrature->Rule.Debug=Debug;
      }

    HeterogenousTM(std::string bpn): Temperature(), Periodic_boundary_condition(false), Debug(false)
      {
      Quadrature = new quadrature_type();
      Quadrature->Rule.Save_calculations=true;
      Quadrature->Rule.Load_calculations=true;
      Quadrature->Rule.Debug=Debug;
      base(bpn);
      }
      
    inline void debug(bool d) 
      {
      Debug=d;
      Quadrature->Rule.Debug=d;
      }
      
    inline void matrix_base_directory(std::string mbase)
      {
      MatrixDir=mbase;
      Quadrature->Rule.Base_directory=mbase;
      if (Debug) std::cerr << "New base directory for matrices [" << mbase << "]"  << std::endl;
      }
      
    inline void verbose(bool v) {Verbose=v;}

    template<class _St>
    inline std::string c_matrix(_St& nm) {return (std::string)nm;}

    template<class _St>
    inline std::string a_matrix(_St& nm) {return (std::string)nm+"*";}

    inline std::string matrix_name(std::string tp, std::string nm)
      {
      if (tp == "A") return a_matrix(nm);
      if (tp == "C") return c_matrix(nm);
      return std::string();
      }

    inline std::pair<std::string,std::string> matrix_name(std::string mm)
      {
      std::string tp,bp;
      if (mm.find("*") != std::string::npos) 
        {
        tp="A";
        bp=mm.substr(0,mm.length()-1);
        }
      else                                   
        {
        tp="C";
        bp=mm;
        }
      return std::pair<std::string,std::string>(tp,bp);
      }

    /// \brief Adds a new TransferMatrix to the map.
    typename tm_map_type::iterator new_transfer_matrix(const BasePairNeighbours<>& bpn) ///< The base pair neighbours name
      {
      typename tm_map_type::iterator pos=TM_map.find((std::string)bpn);
      if (pos == TM_map.end())
        {
        TM_map[(std::string)bpn]=transfer_matrix_type(bpn);
        delete TM_map[(std::string)bpn].quadrature;
        TM_map[(std::string)bpn].quadrature=Quadrature;
        Range_map[c_matrix(bpn)]=Range<double>();
        //looping over A matrices seems not to work yet
        //Range_map[a_matrix(bpn)]=Range<double>();
        pos=TM_map.find((std::string)bpn);
        if (pos != TM_map.end()) return pos;
        else return TM_map.end();
        }
      return pos;
      }

    /// \brief Returns a TransferMatrix identified by the base pair neighbours name.
    transfer_matrix_type* operator()(std::string bpn) ///< The base pair neighbours name
      {return &(TM_map.find(bpn)->second);}

    /// \brief Defines the base on which the expansion takes place.
    void base(const BasePairNeighbours<>& bpn)
      {
      BaseName=(std::string)bpn;
      Base=TM_map.find((std::string)bpn);
      if (Base == TM_map.end()) 
        {
        new_transfer_matrix(bpn);
        Base=TM_map.find((std::string)bpn);
        }
      }

    /// \brief Adds a new basepair .
    void add_basepair(const BasePairNeighbours<>& bpn)
      {
      typename tm_map_type::iterator itm;
      itm=TM_map.find((std::string)bpn);

      if (itm == TM_map.end()) 
        {
        itm=new_transfer_matrix(bpn);
        itm->second.Hamiltonian_ID=Base->second.Hamiltonian_ID;
        itm->second.get_parameters(parameter_map);
        itm->second.eigenvalue_cutoff(Base->second.Eigenvalue_cutoff);
        itm->second.limits(Base->second.Lower_limit,Base->second.Upper_limit,Base->second.Integration_steps);
        itm->second.temperature(Temperature);
        }
      }

    void select_hamiltonian(std::string hamiltonian_name)
      {
      typename tm_map_type::iterator it;
      for(it=TM_map.begin(); it != TM_map.end(); ++it)
        it->second.select_hamiltonian(hamiltonian_name);
      }

    inline std::string hamiltonian(void)
      {
      return Base->second.hamiltonian();
      }

    /// \brief Reads one parameter file and assigns to all maps
    inline bool get_parameters(std::string par_file)
      {
      parameter_map.read_from(par_file);
      ParameterIdName=parameter_map.Identify;
      typename tm_map_type::iterator it;
      get_parameters();
      return true;
      }


    /// \brief Reads a deque of parameter files and assigns to all maps
    inline bool get_parameters(string_deque_type &par_file_deque)
      {
      typename string_deque_type::iterator pit;
      for(pit=par_file_deque.begin(); pit != par_file_deque.end(); pit++)
        parameter_map.read_from(*pit);
      ParameterIdName=parameter_map.Identify;
      get_parameters();
      return true;
      }

    inline bool get_parameters(void)
      {
      typename tm_map_type::iterator it;
      for(it=TM_map.begin(); it != TM_map.end(); ++it)
        it->second.get_parameters(parameter_map);
      return true;
      }

    void limits(value_type l, ///< Lower integration limit, usually negative 
                value_type u, ///< Upper integration limit
                size_t s)     ///< Number of integration points
      {
      typename tm_map_type::iterator it;
      for(it=TM_map.begin(); it != TM_map.end(); ++it)
        it->second.limits(l,u,s);
      }

    void clear_all_matrices(void)
      {
      C_matrix_map.clear();
      Delta_matrix_map.clear();
      ADelta_matrix_map.clear();
      Y.clear();
      Eigenvector.clear();
      Eigenvalue.clear();
      Lambda_matrix.clear();      
      }

    void temperature(double T) 
      {
      if (Temperature != T)
        {
        Temperature=T;
        // Whenever the temperature changes we need new matrices.
        clear_all_matrices();

        // Assign the temperature to all known TransferMatrix classes
        typename tm_map_type::iterator it;
        for(it=TM_map.begin(); it != TM_map.end(); ++it)
          it->second.temperature(T) ;
        }
      }

    value_type temperature(void) {return Temperature;}

    /// \brief Maintained for compatibility only.
    ///
    /// This method does nothing, it is kept for compatibility with
    /// certain Action classes.
    void calculate_eigenvalue_only(bool ceo) {}

    void eigenvalue_cutoff(size_t c) 
      {
      typename tm_map_type::iterator it;
      for(it=TM_map.begin(); it != TM_map.end(); ++it)
        it->second.eigenvalue_cutoff(c) ;
      }

    size_t eigenvalue_cutoff(void) 
      {
      size_t N=Base->second.Integration_steps;
      if (Base->second.Eigenvalue_cutoff > (size_t)0) 
        N=Base->second.Eigenvalue_cutoff;
      return N;
      }


    friend std::ostream& operator<<(std::ostream &out, const HeterogenousTM& tm)
      {
      typename tm_map_type::const_iterator it;
      for(it=tm.TM_map.begin(); it != tm.TM_map.end(); ++it)
        out << it->second;
      return out;
      }

    std::string matrix_filename(std::string mat_name, bp_id_type bp_name=bp_id_type())
      {
      std::stringstream filename;
      filename << MatrixDir << "/" << ParameterIdName
               << "-" << Base->first
               << "-" << Temperature << "-" << mat_name;
      if (bp_name != bp_id_type()) filename << "." << bp_name;
      return filename.str();
      }

    /// \brief Stores matrices into the map and saves into a directory.
    ///
    /// \attention If you don't want to store the matrix in a file
    /// leave MatrixDir empty.
    void save_matrix(const matrix_type& mat,
                     std::string mat_name, bp_id_type bp_name=bp_id_type())
      {
      if (!MatrixDir.empty())
        {
        boost::filesystem::create_directory(MatrixDir);
        std::ofstream outfile;
        std::string filename;
        filename=matrix_filename(mat_name,bp_name);
        outfile.open(filename.c_str());
        outfile << *this << std::endl;
        outfile << mat;
        outfile.close();
        Quadrature->Rule.Base_directory=MatrixDir;
        }
      if ((mat_name == "C") || (mat_name == "A"))
        {
        if (C_matrix_map.find(matrix_name(mat_name,bp_name)) == C_matrix_map.end())
          C_matrix_map.insert(typename matrix_map_type::value_type(matrix_name(mat_name,bp_name),mat));
        else
          C_matrix_map[matrix_name(mat_name,bp_name)]=mat;
        }
      }

    void save_matrices(void)
      {
      if (!MatrixDir.empty())
        {
        boost::filesystem::create_directory(MatrixDir);
        Quadrature->Rule.Base_directory=MatrixDir;
        typename matrix_map_type::iterator mm;
        for (mm=C_matrix_map.begin(); mm != C_matrix_map.end(); ++mm)
          {
          std::ofstream outfile;
          std::string filename;
          std::pair<std::string,std::string> mp=matrix_name(mm->first);
          filename=matrix_filename(mp.first,mp.second);
          outfile.open(filename.c_str());
          outfile << mm->second;
          outfile.close();
          }
        }
     }

    void retrieve_matrices(void)
      {
      size_t N=eigenvalue_cutoff();
      matrix_type C(N,N);
      C.lock_size();
      std::string filename;
 
      if (Y.size() == (size_t)0)
        {
        filename=matrix_filename("Y");
        if (boost::filesystem::exists(filename)) 
          {
          if (Debug) std::cerr << "Reading " << filename << std::endl;
          std::ifstream in(filename.c_str());
          Y.resize(N,N); Y.lock_size();
          in >> Y;
          }
        else
          {
          if (Verbose) std::cerr << filename << " not found, calculating ...";
          calculate_matrix_Y();
          if (Verbose) std::cerr << std::endl;
          }
        }

      typename tm_map_type::iterator it;
      for(it=TM_map.begin(); it != TM_map.end(); ++it)
        {
        if (C_matrix_map.find(it->first) == C_matrix_map.end())
          {
          filename=matrix_filename("C",it->first);
          if (boost::filesystem::exists(filename)) 
            {
            if (Debug) std::cerr << "Reading " << filename << std::endl;
            std::ifstream in(filename.c_str());
            in >> C;
            C_matrix_map.insert(typename matrix_map_type::value_type(c_matrix(it->first),C));
            }
          else 
            {
            if (Verbose) std::cerr << filename << " not found, calculating ...";
            calculate_matrix_C(it);
            if (Verbose) std::cerr << std::endl;
            }
          }
        if (C_matrix_map.find(a_matrix(it->first)) == C_matrix_map.end())
          {
          filename=matrix_filename("A",it->first);
          if (boost::filesystem::exists(filename)) 
            {
            if (Debug) std::cerr << "Reading " << filename << std::endl;
            std::ifstream in(filename.c_str());
            in >> C;
            C_matrix_map.insert(typename matrix_map_type::value_type(a_matrix(it->first),C));
            }
          else
            {
            if (Verbose) std::cerr << filename << " not found, calculating ...";
            calculate_matrix_A(it);
            if (Verbose) std::cerr << std::endl;
            }
          }
        }
      generate_delta_matrices();
      }

    void calculate_eigensystem(void)
      {
      Base->second.Recalculate_diagonalization=true;
      Base->second.calculate_eigenvalue_only(false);
      Base->second.diagonalize();
      Eigenvector=Base->second.eigen_engine.Eigenvector;
      Eigenvalue=Base->second.eigen_engine.Eigenvalue;
      save_matrix(Eigenvector,"eigenvectors");
      save_matrix(Eigenvalue,"eigenvalues");
      }

    void retrieve_eigensystem(void)
      {
      std::string filename=matrix_filename("eigenvectors");
      if (boost::filesystem::exists(filename)) 
        {
        if (Debug) std::cerr << "Reading " << filename  << std::endl;
        std::ifstream in(filename.c_str());
        in >> Eigenvector;
        in.close();
        filename=matrix_filename("eigenvalues");
        if (Debug) std::cerr << "Reading " << filename  << std::endl;
        in.open(filename.c_str());
        in >> Eigenvalue;
        in.close();
        }
      else
        {
        if (Verbose) std::cerr << filename << " not found, calculating ...";
        calculate_eigensystem();
        if (Verbose) std::cerr << std::endl;
        }
      }

    void calculate_matrix_C(typename tm_map_type::iterator it)
      {
      if (Eigenvector.size() == (size_t)0) retrieve_eigensystem();

      size_t N=eigenvalue_cutoff();
      matrix_type C(N,N);
      if (it == Base) C.diagonal(Eigenvalue.submatrix(0,0,1,N));
      else C=it->second.C_matrix(Eigenvector);
      save_matrix(C,"C",it->first);
      }


    void calculate_matrix_C(std::string bp)
      {
      typename tm_map_type::iterator it=TM_map.find(bp);
      if (it != TM_map.end()) calculate_matrix_C(it);
      }


    void calculate_matrix_A(typename tm_map_type::iterator it)
      {
      if (Eigenvector.size() == (size_t)0) retrieve_eigensystem();
      size_t N=eigenvalue_cutoff();
      matrix_type A(N,N);
      A=it->second.A_matrix(Eigenvector);
      save_matrix(A,"A",it->first);
      }


    void calculate_matrix_Y()
      {
      if (Eigenvector.size() == (size_t)0) retrieve_eigensystem();
      Y=Base->second.Y_matrix(Eigenvector);
      save_matrix(Y,"Y");
      }

    void diagonalize(void)
      {
      if (Base->second.Recalculate_diagonalization)
        {
        C_matrix_map.clear();
        calculate_eigensystem();
        calculate_matrix_C(Base);
        calculate_matrix_A(Base);
        calculate_matrix_Y();
     
        typename tm_map_type::iterator it;
        for(it=TM_map.begin(); it != TM_map.end(); ++it)
          {
          if (it != Base)
            {
            calculate_matrix_C(it);
            calculate_matrix_A(it);
            }
          }
        }
      generate_delta_matrices();
      }

    void generate_delta_matrices(void)
      {
      Delta_matrix_map.clear();
      ADelta_matrix_map.clear();
      Lambda_matrix=C_matrix_map[c_matrix(BaseName)];
      typename tm_map_type::iterator it;
      for(it=TM_map.begin(); it != TM_map.end(); ++it)
        {
        matrix_type D(Lambda_matrix.Rows,Lambda_matrix.Columns);
        if (it != Base)
          D=C_matrix_map[c_matrix(it->first)]-Lambda_matrix;
        else D.zero();
        Delta_matrix_map.insert(typename matrix_map_type::value_type(it->first,D));
        D=C_matrix_map[a_matrix(it->first)]-Lambda_matrix;
        ADelta_matrix_map.insert(typename matrix_map_type::value_type(it->first,D));        
        }
      }


    /// \brief Sets the sequence and analyses it.
    ///
    /// By calling this method the class determines which matrices need
    /// to be generated.
    void sequence(const std::string& seq, const std::string& comp=std::string())
      {
      // Only used for CG-contents analysis, the base-pairs are symmetry reduced
      // i.e., TA becomes AT and so on.
      Equivalent_duplex.clear();
      if(comp != std::string())
        Equivalent_duplex=Duplex<>(seq+"/"+comp,true,BasePair<>::not_context_dependent);
      else 
        Equivalent_duplex=Duplex<>(seq,true,BasePair<>::not_context_dependent);
 
      // Only used for CG-contents analysis
      Duplex_set.clear();
      Duplex_set.Use_cheapest_symmetry=true;
      Duplex_set.analyse(Equivalent_duplex);

      // This is the critical part, the duplex is generated without touching the
      //basepairs, i.e., they are not symmetry reduced.
      //Duplex stores a deque of base-pairs
      Duplex<> exact_duplex;
      if(comp != std::string())
        exact_duplex=Duplex<>(seq+"/"+comp,false,BasePair<>::context_dependent);
      else 
        exact_duplex=Duplex<>(seq,false,BasePair<>::context_dependent);

      // Equivalent_neighbours is of type NeighbourSequence which stores a deque of
      // nearest neighbours
      Equivalent_neighbours.clear();
      Equivalent_neighbours=exact_duplex;
      Equivalent_neighbours.reduce_to_smallest_symmetry();

      Neighbours_set.clear();
      Neighbours_set.analyse(Equivalent_neighbours);


      if (Debug)
        {
        std::cerr << "Debug       : " << seq << " " << comp << std::endl
                  << "Exact Duplex: " << exact_duplex << std::endl
                  << "Equiv. neigh: " << Equivalent_neighbours << std::endl
                  << "Neighb.  set: " << Neighbours_set << std::endl;
        }

/*
      std::cout << std::endl << seq << std::endl << comp;
      std::cout << std::endl << exact_duplex;
      std::cout << std::endl << Equivalent_neighbours;
      std::cout << std::endl << Duplex_set << std::endl;
      std::cout << std::endl << Neighbours_set << std::endl;
*/
      typename neighbours_ref_type::const_iterator it;
      for(it=Neighbours_set.begin(); it != Neighbours_set.end(); ++it)
        {

        if (TM_map.find(it->first) == TM_map.end())
          {
          //Context_dependence now need to be switched off
          BasePairNeighbours<> Bpn=it->first;
          Bpn.first.Context_Dependence=false;
          Bpn.second.Context_Dependence=false;
          add_basepair(Bpn);
          }
        }
      }

    value_type partition_function(void)
      {
      size_t N=eigenvalue_cutoff();
      matrix_type Z(N,N);
      Z.diagonal(1.0);
      NeighbourSequence<>::const_iterator it;
      size_t i;
      for (it=Equivalent_neighbours.begin(), i=1; it != Equivalent_neighbours.end(); ++it, ++i)
        {
        matrix_type C=C_matrix_map[c_matrix(*it)];
        if (!Periodic_boundary_condition && (i==Equivalent_neighbours.size()-1)) 
          C=C_matrix_map[a_matrix(*it)];
        if (C.size() == size_t()) 
          {std::cout << "No C matrix for " << (std::string)(*it) << std::endl; exit(1);}
        Z=Z*C;
        }

      return Z.trace();
      }

    std::valarray<value_type> average_y(void)
      {
      size_t N=eigenvalue_cutoff();
      matrix_type Z(N,N), C(N,N);

      std::valarray<value_type> res(Equivalent_neighbours.size());
      std::vector<matrix_type> F(Equivalent_neighbours.size()+1);
      std::vector<matrix_type> B(Equivalent_neighbours.size()+1);

      NeighbourSequence<>::const_iterator it;
      size_t i;
      Z.diagonal(1.0);
      F[0]=Z;
      for (it=Equivalent_neighbours.begin(), i=1; it != Equivalent_neighbours.end(); ++it, ++i)
        {
        if (!Periodic_boundary_condition && (i==Equivalent_neighbours.size())) 
          C=C_matrix_map[a_matrix(*it)];
        else
          C=C_matrix_map[c_matrix(*it)];
        if (C.size() == size_t()) 
          {std::cout << "No C matrix for " << (std::string)(*it) << std::endl; exit(1);}
        F[i]=F[i-1]*C;
        }

      value_type z=F[Equivalent_neighbours.size()].trace();

      Z.diagonal(1.0);
      B[Equivalent_neighbours.size()]=Z;
      NeighbourSequence<>::reverse_iterator bit;
      for (bit=Equivalent_neighbours.rbegin(), i=Equivalent_neighbours.size()-1; bit != Equivalent_neighbours.rend(); ++bit, --i)
        {
        if (!Periodic_boundary_condition && (i==(Equivalent_neighbours.size()-1))) 
          C=C_matrix_map[a_matrix(*bit)];
        else
          C=C_matrix_map[c_matrix(*bit)];
        if (C.size() == size_t()) 
          {std::cout << "No C matrix for " << (std::string)(*bit) << std::endl; exit(1);}
        B[i]=C*B[i+1];
        }

      Z=Y*B[0];
      res[0]=Z.trace()/z;

      for (i=1; i < Equivalent_neighbours.size(); ++i)
        {
        Z=F[i]*Y*B[i];
        res[i]=Z.trace()/z;
        }
      

      return res;
      }

    inline std::valarray<value_type> expand_partition_function(void)
      {
      std::vector<matrix_type> Gamma(Equivalent_neighbours.size()+1);
      NeighbourSequence<>::const_iterator it=Equivalent_neighbours.begin();

      Gamma[1]=Lambda_matrix;
      if (it != Equivalent_neighbours.end())
        Gamma[0]=Delta_matrix_map[(std::string)(*it)];
      else
        {
        std::cerr << "Void Equivalent_neighbours in " << __FILE__ << " line " << __LINE__ << std::endl;
        exit(1);
        }
      size_t i;
      for (++it, i=1; it != Equivalent_neighbours.end(); ++it, ++i)
        {
        matrix_type Delta;
        if (!Periodic_boundary_condition && (i==Equivalent_neighbours.size()-1)) 
          Delta=ADelta_matrix_map[(std::string)(*it)];
        else
          Delta=Delta_matrix_map[(std::string)(*it)];
        Gamma[i+1]=Gamma[i]*Lambda_matrix;
        for (size_t j=i; j > 0; --j)
          Gamma[j]=Gamma[j]*Delta+Gamma[j-1]*Lambda_matrix;
        Gamma[0]=Gamma[0]*Delta;
        }

      std::valarray<value_type> res(Equivalent_neighbours.size()+1);
      for (i=0; i < Equivalent_neighbours.size()+1; ++i)
        res[i]=Gamma[i].trace();
      return res;
      }

    inline value_type melting_index(void)
      {
      std::valarray<value_type> Z_Lambda=expand_partition_function();
      std::valarray<value_type> xi(Z_Lambda.size());

      for(size_t i=0; i < xi.size(); ++i) xi[i]=static_cast<value_type>(i);

      std::valarray<bool> positive=Z_Lambda > value_type();

      std::valarray<value_type> f=Z_Lambda[positive];
      std::valarray<value_type> x=xi[positive];

      value_type s2pi=sqrt(2*M_PI);

      value_type  sigma1=1.0/(f.max()*s2pi);
      value_type  sigma2=sqrt(x.max()-x.min())/2.0;

      std::valarray<value_type> xm=x[f == f.max()];
      value_type  imu = xm[0];

      value_type  best_mu=0.0, lowest_std=std::numeric_limits<value_type>::max();
      for (value_type mu=imu-1.0; mu <= imu+1.0; mu+=0.001)
        {
        std::valarray<value_type> gs=exp(-pow(x-mu,2.0)/(2*pow(sigma2,2.0)))/(sigma1*s2pi);
        value_type df=std::valarray<value_type>(pow(f-gs,2.0)).sum();
        if ((df < lowest_std))
          {
          lowest_std=df; 
          best_mu=mu;
          }
        }
       return best_mu;
       }

    /// \brief Prepares the matrices for the minimization process
    ///
    /// Range_map is initialized in the constructor. First we take
    /// out the expansion base which is constant during the minimization
    /// process.
    inline void setup_loop_matrices(Range<double>& rg, std::deque<std::string>& constant)
      {
      Range_map.erase(c_matrix(Base->first)); 
      Range_map.erase(a_matrix(Base->first));

      if (constant.size() > 0)
        {
        typename std::deque<std::string>::const_iterator ct;
        for(ct=constant.begin(); ct != constant.end(); ++ct)
          {
          Range_map.erase(c_matrix(*ct)); 
          Range_map.erase(a_matrix(*ct));
          }
        }

      typename range_map_type::iterator r;
      for(r=Range_map.begin(); r != Range_map.end(); ++r)
        {
        r->second=rg;
        C_matrix_map[r->first][0]=r->second.current();
        }

      }

    inline void matrix2vector(vector_type& vec)
      {
      vec.resize(Range_map.size());
      typename range_map_type::iterator r;
      size_t n;
      for(n=0, r=Range_map.begin(); r != Range_map.end(); ++r, ++n)
        vec[n]=C_matrix_map[r->first][0];
      }

    inline void set_loop_matrices(const vector_type& v)
      {
      typename range_map_type::iterator r;
      size_t n;
      for(n=0, r=Range_map.begin(); r != Range_map.end(); ++r, ++n)
        C_matrix_map[r->first][0]=v[n];
      }

   inline void print_current(std::ostream& out)
      {
      typename range_map_type::iterator rm;
      for (rm=Range_map.begin(); rm != Range_map.end(); ++rm)
         out << rm->second.current() << " ";
      out << std::endl;
      }

    inline void print_matrix_head(std::ostream& out)
      {
      typename matrix_map_type::iterator mm;
      for (mm=C_matrix_map.begin(); mm != C_matrix_map.end(); ++mm)
        {
        std::string mat_name=mm->first;
        if (Range_map.find(mm->first) != Range_map.end()) mat_name="["+mat_name+"]";
        out << std::setw(10) << mat_name << " ";
        }
      out << std::endl;
      }

    inline void print_matrix(std::ostream& out)
      {
      typename matrix_map_type::iterator mm;
      for (mm=C_matrix_map.begin(); mm != C_matrix_map.end(); ++mm)
         out << std::setw(10) << mm->second[0] << " ";
      }

    inline value_type approx_melting_index(void)
      {
      value_type omega=value_type(), lambda=Lambda_matrix[0];
      neighbours_ref_type::const_iterator it;
      for (it=Neighbours_set.begin(); it != Neighbours_set.end(); ++it)
        {
        if (it->second > value_type())
          omega+=it->second*lambda/C_matrix_map[c_matrix(it->first)][0];
        }
      // For non-periodic BC we remove one C matrix and add the corresponding A matrix
      if (!Periodic_boundary_condition)
        {
        NeighbourSequence<>::reverse_iterator bit=Equivalent_neighbours.rbegin();
        omega-=lambda/C_matrix_map[c_matrix(*bit)][0];
        omega+=lambda/C_matrix_map[a_matrix(*bit)][0];
        }
      return omega;
      }

    inline value_type approx_melting_index_loop(void)
      {
      (*Loop_matrix_map)[Base->first][0]=1.0;
      value_type omega=value_type();
      neighbours_ref_type::const_iterator it;
      for (it=Neighbours_set.begin(); it != Neighbours_set.end(); ++it)
        {
        if (it->second > value_type())
          omega+=it->second*(*Loop_matrix_map)[(std::string)(it->first)][0];
        }
      return omega;
      }

    inline value_type cg_content(void)
      {
      value_type cg=0;
      Duplex<>::iterator it=Equivalent_duplex.begin();
      for (it=Equivalent_duplex.begin(); it != Equivalent_duplex.end(); ++it)
        {
        if ((std::string)(*it) == "CG" || (std::string)(*it) == "GC")
          cg+=1.0;
        }
      //std::cout << Sequence << " " << cg << std::endl;
      return cg/Equivalent_duplex.size();
      }

    };
  };

#endif
