// $Id: ModelPotentials.h 698 2012-11-13 16:24:16Z ge $
/// \file ModelPotentials.h
/// \brief contains the model Potentials function class
///
/// $Revision: 698 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_EXP_MODELPOTENTIALS_H
#define GBC_EXP_MODELPOTENTIALS_H "$Id: ModelPotentials.h 698 2012-11-13 16:24:16Z ge $"
#include "BasePair.h"
#include "BasePairNeighbours.h"
#include "Potentials.h"
using namespace gbc;
namespace gbc_exp
  {
  template<class _Tp=double>
  /// \brief Morse potential
  ///
  /// \f$ V(y)=D\left(e^{-y/\lambda_a}-1\right)^2 \f$
  class Morse: public BasePotential<_Tp>
    {
    public:
      typedef BasePotential<_Tp> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      typedef _Tp value_type;
      typedef std::valarray<value_type> vector_type;

      BasePair<> bp;
      value_type lambda; ///< \f$ \lambda_a \f$
      value_type D;      ///< \f$ D \f$, depth of the potential

    Morse(void) : bp(), lambda(), D() 
      {
      base_type::name("morse");
      base_type::prefix(std::string());
      }

    Morse(const BasePair<>& b) : bp(b), lambda(), D() 
      {
      base_type::name("morse");
      bp.reduce_to_smallest_symmetry();
      base_type::prefix((std::string)bp);
      }

    /// \brief Retrieve the parameter from the map
    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      this->define_parameter(pm,"lambda",lambda);
      this->define_parameter(pm,"D",D);
      return true;
      }

    inline bool parameters(const Morse<value_type>& ms)
      {
      bool modified=(lambda != ms.lambda) || (D != ms.D);
      lambda=ms.lambda; 
      D=ms.D;
      return modified;
      }

    inline virtual value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      return D*pow(exp(-x/lambda)-1.0,2);
      }
    };

  template<class _Tp=double>
  /// \brief The solvent potential.
  ///
  /// \f$ V_s(y)=-f_sD\left[\tanh(y/\lambda_s)+g_s\right] \f$
  /// \attention This also uses parameters from the Morse Potential
  class Solvent: public BasePotential<_Tp>
    {
    public:
      typedef BasePotential<_Tp> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      typedef _Tp value_type;
      typedef std::valarray<value_type> vector_type;

      BasePair<> bp;
      Morse<value_type> morse; ///< The Morse potential, for use of \f$ D \f$
      value_type f_s;          ///< \f$ f_s \f$
      value_type lambda;       ///< \f$ \lambda_s \f$ 
      value_type eq_sol;       ///< equilibrium distance
      value_type sign_sol;     ///< the sign \f$ g_s \f$ 

    Solvent(void) : bp(), morse(), f_s(), lambda(), eq_sol(), sign_sol() 
      {
      base_type::name("solvent");
      base_type::prefix(std::string());
      }

    Solvent(const BasePair<>& b) : bp(b), morse(b), f_s(), lambda(), eq_sol(), sign_sol() 
      {
      base_type::name("solvent");
      bp.reduce_to_smallest_symmetry();
      base_type::prefix((std::string)bp);
      }

    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      this->define_parameter(pm,"f_s",f_s);
      this->define_parameter(pm,"lambda",lambda);
      this->define_parameter(pm,"eq_sol",eq_sol);
      this->define_parameter(pm,"sign_sol",sign_sol);
      morse.get_parameters(pm);
      base_type::ParMap.insert(morse.ParMap.begin(),morse.ParMap.end());
      return true;
      }

    inline bool parameters(const Solvent<value_type>& sol)
      {
      bool modified=false;
      modified=modified || morse.parameters(sol.morse);
      modified=modified || (f_s !=sol.f_s) || (lambda != sol.lambda)
               || (eq_sol != sol.eq_sol) || (sign_sol != sol.sign_sol);
      f_s=sol.f_s; lambda=sol.lambda; eq_sol=sol.eq_sol; sign_sol=sol.sign_sol;
      return modified;
      }

    inline virtual value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      return -f_s*morse.D*(tanh((x+eq_sol)/lambda)+sign_sol);
      }
    };

  template<class _Tp=double>
  class MorseSolvent: public CmbPotential<Morse<_Tp>,Solvent<_Tp> >
    {
    typedef  CmbPotential<Morse<_Tp>,Solvent<_Tp> > base_type;
    public:
      typedef _Tp value_type;
      MorseSolvent(void): CmbPotential<Morse<_Tp>,Solvent<_Tp> >() 
       {} 
      MorseSolvent(const BasePair<>& b): CmbPotential<Morse<_Tp>,Solvent<_Tp> >(b) 
       {} 
    };


  template<class _Tp=double>
  class HarmonicStacking: public BasePotential<_Tp>
    {
    public:
      typedef BasePotential<_Tp> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      typedef _Tp value_type;
      typedef std::valarray<value_type> vector_type;

      BasePairNeighbours<> bpn;
      value_type theta, k;

    HarmonicStacking(void) : bpn(), theta(), k()
      {
      base_type::name("harmonic");
      base_type::prefix(std::string());
      }

    HarmonicStacking(const BasePairNeighbours<>& b) : bpn(b), theta(), k()
      {
      base_type::name("harmonic");
      bpn.reduce_to_smallest_symmetry();
      base_type::prefix((std::string)bpn);
      }

    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      this->define_parameter(pm,"theta",theta);
      this->define_parameter(pm,"k",k);
      return true;
      }

    inline bool parameters(const HarmonicStacking<value_type>& hs)
      {
      bool modified=(theta != hs.theta) || (k != hs.k);
      theta=hs.theta; 
      k=hs.k;
      return modified;
      }

    inline virtual value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      return 0.5*k*(pow(x,2)-2.0*x*y*cos(theta)+pow(y,2));
      }
    };

  template<class _Tp=double>
  class AnharmonicStacking: public BasePotential<_Tp>
    {
    public:
      typedef BasePotential<_Tp> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      typedef _Tp value_type;
      typedef std::valarray<value_type> vector_type;

      BasePairNeighbours<> bpn;
      value_type rho, alpha;
      HarmonicStacking<value_type> harmonic;

    AnharmonicStacking(void) : bpn(), rho(), alpha(), harmonic()
      {
      base_type::name("anharmonic");
      base_type::prefix(std::string());
      }

    AnharmonicStacking(const BasePairNeighbours<>& b) : bpn(b), rho(), alpha(), harmonic(b)
      {
      base_type::name("anharmonic");
      bpn.reduce_to_smallest_symmetry();
      base_type::prefix((std::string)bpn);
      }

    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      this->define_parameter(pm,"rho",rho);
      this->define_parameter(pm,"alpha",alpha);
      harmonic.get_parameters(pm);
      base_type::ParMap.insert(harmonic.ParMap.begin(),harmonic.ParMap.end());
      return true;
      }

    inline bool parameters(const AnharmonicStacking<value_type>& ast)
      {
      bool modified=harmonic.parameters(ast.harmonic);
      modified = modified || (rho != ast.rho) || (alpha != ast.alpha);
      rho=ast.rho; alpha=ast.alpha;
      return modified;
      }

    inline virtual value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {return (1.0+rho*exp(-alpha*(x+y)))*harmonic.evaluate_xy(x,y);}
    };

  template<class _Tp=double>
  class FiniteEnthalpyStacking: public BasePotential<_Tp>
    {
    public:
      typedef BasePotential<_Tp> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      typedef _Tp value_type;
      typedef std::valarray<value_type> vector_type;

      BasePairNeighbours<> bpn;
      value_type DeltaH, b, kb, C;

    FiniteEnthalpyStacking(void) : bpn(), DeltaH(), b(), kb()
      {
      base_type::name("finite_enthalpy");
      base_type::prefix(std::string());
      }

    FiniteEnthalpyStacking(const BasePairNeighbours<>& bp) : bpn(bp), DeltaH(), b(), kb()
      {
      base_type::name("finite_enthalpy");
      bpn.reduce_to_smallest_symmetry();
      base_type::prefix((std::string)bpn);
      }

    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      this->define_parameter(pm,"DeltaH",DeltaH);
      this->define_parameter(pm,"b",b);
      this->define_parameter(pm,"kb",kb);
      this->define_parameter(pm,"C",C); //see joyeux09 Eq. 2.1, originally C=2.0
      return true;
      }

    inline bool parameters(const FiniteEnthalpyStacking<value_type>& ast)
      {
      bool modified=false;
      modified = modified || (DeltaH != ast.DeltaH) || (b != ast.b) || (kb != ast.kb) 
                          || (C != ast.C);
      DeltaH=ast.DeltaH; b=ast.b; kb=ast.kb; C=ast.C;
      return modified;
      }

    inline virtual value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      value_type xy2=pow(x-y,2);
      return DeltaH/C*(1.0-exp(-b*xy2))+kb*xy2;
      }
    };

  template<class _Tp=double>
  class Yukawa: public BasePotential<_Tp>
    {
    public:
      typedef BasePotential<_Tp> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      typedef _Tp value_type;
      typedef std::valarray<value_type> vector_type;
      value_type k, lambda;

    Yukawa(std::string p=std::string()) : k(), lambda()
      {
      base_type::name("yukawa");
      base_type::prefix(p);
      }

    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      this->define_parameter(pm,"k",k);
      this->define_parameter(pm,"lambda",lambda);
      return true;
      }

    inline bool parameters(const Yukawa<value_type>& hs)
      {
      bool modified=(lambda != hs.lambda) || (k != hs.k);
      lambda=hs.lambda; 
      k=hs.k;
      return modified;
      }

    inline virtual value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      return k*exp(-fabs(x)/lambda)/fabs(x);
      }
    };
  };

#endif
