// $Id: Nucleotide.h 628 2011-08-31 15:42:16Z ge $
// Copyright 2006 Gerald Weber
/// \file Nucleotide.h
/// \brief Contains the definition of the class Nucleotide
///
/// $Revision: 628 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_NUCLEOTIDE_H
#define GBC_NUCLEOTIDE_H "$Id: Nucleotide.h 628 2011-08-31 15:42:16Z ge $"
#include "JobControl.h"
#include "SequencePosition.h"
#include "NucleotideDictionary.h"
#include <fstream>

namespace gbc 
  {
  /// \brief Basic definition of a Nucleotide.
  ///
  /// The basic class which holds the information about a nucleotide in a sequence.
  /// _InternalTp (default=char) is the way the information will be actually stored.
  ///
  /// Nucleotide uses an internal dictionary to convert from symbols to it's internal
  /// representation given by _InternalTp. Some are predefined, see file Nucleotide.cpp.
  template<class _InternalTp=char>
  class Nucleotide 
    {
    public:
      typedef NucleotideDictionary<_InternalTp> DictionaryType;
      _InternalTp representation;         ///< the internal representation
      bool Good;                          ///< Flags which tells us if Nucleotide has read a valid value.
      static DictionaryType  dictionary;  ///< Dictionary which translates file symbols to internal representation
	                                  ///< This object is static because it makes sense that all objects of type Nucleotide use the same
                                          ///< conversion rules. See file NucleotideDictionary.cpp for predefined dictionaries.
      SequencePosition Sequence_position; ///< Position of this Nucleotide in a file

      /// \brief Void onstructor.
      Nucleotide(void): representation(),Good(false)
	{}

      /// \brief Constructor which initializes with a specific dictionary.
      Nucleotide(const char cr): representation(),Good(false)
	{
        (*this)=cr;
	};


      /// \brief Constructor which initializes with a specific dictionary.
      ///
      /// Note that the dictionary needs to be fully specified beforehand.
      Nucleotide(DictionaryType &dict) ///< the object of type DictionaryType
        : representation(), Good(false)
	{
        dictionary=dict;
        }

      /// Returns the symbol corresponding to its internal representation.
      inline char symbol (void) const {return dictionary.invert(representation);};

      /// Returns the symbol corresponding to its internal representation.
      inline char symbol (void) {return dictionary.invert(representation);};

      /// \brief Returns the content of the internal representation.
      ///
      /// E.g., if _InternalTp=int then you can write (int)nc where nc is of type Nucleotide.
      operator _InternalTp() {return representation;}

      template<class _AddTp>
      /// \brief Assigns _AddTp into Nucleotide<_InternalTp>.
      ///
      /// There must be a way to convert _AddTp into _InternalTp, i.e.,
      /// _InternalTp=_AddTp must be possible.
      /// \attention The flag Good will always be true if not set by the convertion _AddTp=_InternalTp.
      Nucleotide& operator = (const _AddTp it) ///< the _InternalTp to be assigned
	{Good=true; representation=it; return *this;}

      /// \brief Allows you to assign a char to Nucleotide.
      ///
      /// This is a specialization of Nucleotide& operator = (_AddTp it)
      /// Sets flag Good=true if assignement was sucessful, i.e., if s is contained in the dictionary.
      Nucleotide& operator = (const char s) ///< the char to translate and assign
	{
        typename DictionaryType::iterator dit=dictionary.find(s);
	Good=dit != dictionary.end();
	if (Good) representation=dit->second;
        else      representation=_InternalTp();
	return *this;
	}

      template<class _InternalTp2>
      /// \brief Assigns Nucleotides with different internal representations.
      ///
      /// This is accomplished by using the symbol() function and assigning as a char.
      /// Because of this there will always be a lookup at the dictionary making this a slow
      /// conversion.
      Nucleotide<_InternalTp>& operator = (const Nucleotide<_InternalTp2>& nt2)
	{
	(*this)=nt2.symbol();
        Sequence_position=nt2.Sequence_position;
        Good=nt2.Good;
	return *this;
	}

      /// \brief Assigns two Nucleotides with the same internal representation.
      ///
      /// This is a specialization of Nucleotide<_InternalTp>& operator = (Nucleotide<_InternalTp2>& nt2)
      /// which takes advantange of the fact that _InternalTp is the same for both and no
      /// dictitonary lookup is necessary.
      Nucleotide<_InternalTp>& operator = (const Nucleotide<_InternalTp>& nt2)
	{
	representation=nt2.representation;
	Sequence_position=nt2.Sequence_position;
        Good=nt2.Good;
	return *this;
	}

      /// \brief Returns the complementary symbol of this Nucleotide.
      /// \attention Function untested.
      inline char complementary_symbol(void) const
        {
        return dictionary.complementary_symbol(symbol());
        }

      /// Converts this Nucleotide to its complementary (\f$A\leftrightarrow T\f$ and \f$C\leftrightarrow G\f$ )
      inline void complementary(void) {representation=dictionary.complementary(representation);}

      /// Tells us if Nucleotide was assigned a valid value.
      inline bool good(void) {return Good;}

      /// Extractor for printing Nucleotide symbols
      friend std::ostream& operator<<(std::ostream &out,const Nucleotide &nt)
	{
	out << nt.symbol();
	return out;
	}

      /// Equal operator returns the operation corresponding to _InternalTp
      inline friend bool operator==(Nucleotide<_InternalTp> nuc1, Nucleotide<_InternalTp> nuc2)
        {
	return nuc1.representation==nuc2.representation;
        }

      /// Less than operator returns the operation corresponding to _InternalTp
      inline friend  bool operator<(Nucleotide nuc1, Nucleotide nuc2)
        {
	return nuc1.representation < nuc2.representation;
	}

      /// Greater than operator returns the operation corresponding to _InternalTp
      inline friend  bool operator>(Nucleotide nuc1, Nucleotide nuc2)
	{
	return nuc1.representation > nuc2.representation;
	}


  }; // end class definition
	

  template<>
  inline char Nucleotide<char>::symbol(void) {return representation;};

  template<class _NucleotideTp>
  inline char NucleotideSymbol(_NucleotideTp nuc) {return nuc.symbol();};

}// end namespace

#endif
