// $Id: Options.h 628 2011-08-31 15:42:16Z ge $
/// \file Options.h
/// \brief contains the classes for options: Option and OptionMap
///
/// $Revision: 628 $
/// \author Gerald Weber <gweberbh@gmail.com>

#ifndef GBC_EXP_OPTIONS_H
#define GBC_EXP_OPTIONS_H "$Id: Options.h 628 2011-08-31 15:42:16Z ge $"

#include "JobControl.h"
#include <deque>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <iterator>
#include "Range.h"

namespace gbc_exp
{

class BaseOption;
template<class _Tp> class Option;

/// \brief Contains the maps of options and arguments.
///
/// This class has several maps of type Option and with the arguments which
/// are read-in from command line. Most functions are static such that they can
/// be called without defining an object.
/// Typically you would have a 
/// \verbatim
///  OptionMap::arguments(argc,argv);
/// \endverbatim
/// somewhere at the beginning of your main(). This reads in the arguments from
/// the command line.
/// Then you define all objects of type Option. When done, you should have a
/// \verbatim
/// OptionMap::finished();
/// \endverbatim
/// which then assigns all arguments to their respective Options.
/// 
/// If you run your program without any options, you should get a brief desription
/// of all options available.
class OptionMap
  {
  public:
  typedef std::map<std::string,BaseOption*>   map_type;
  typedef std::deque<std::string>           arg_deque_type;
  typedef std::map<std::string,std::string> arg_map_type;

  static const bool necessary=true;  ///< to be used to set an option to become necessary.
  static const bool optional=false;  ///< to be used to set an option to be optional.

  static size_t Max_arg_length;       ///< maximal length of arguments, dynamically updated
  static size_t Max_desc_length;      ///< maximal length of descriptions, dynamically updated
  static size_t Max_type_desc_length; ///< maximal length of type descriptions, dynamically updated

  static map_type       all_options;           ///< Holds all options
  static arg_deque_type all_arguments;         ///< All arguments which were provided
  static arg_map_type   interpreted_arguments; ///< All arguments after interpretation.

  static void add_option(std::string arg, BaseOption* bo);

  /// \brief Adds arguments, typically from command-line, to the deque OptionMap::all_arguments
  static void arguments(int argc, char *argv[]);

  /// \brief Takes the argument and brakes it down into its components.
  ///
  /// From "-opt=arg" we brake down into "opt" and "arg". If "arg" has further
  /// structures, e.g. "0:1:2" we leave that for later.
  static void interpretation(std::string arg);

  /// \brief Correlated the arguments with the defined options and assigns the results.
  static void scan_arguments(void);

  static void show_options(int action=0);

  static bool provided(std::string arg);

  /// \brief Finishes reading all arguments and assigns to options.
  static void finished(void);

  static void check_if_all_given(void);

  template<class _Tp> static int type_string(void);

  template<class _Tp> static std::string type_description(void);

  template<class _Tp>
  static void add(std::string arg, Option<_Tp>* so)
    {
    add_option(arg,static_cast<BaseOption*>(so));
    if (OptionMap::type_description<_Tp>().length() > OptionMap::Max_type_desc_length)
      OptionMap::Max_type_desc_length=OptionMap::type_description<_Tp>().length();
    }

  };

/// \brief Base class for all Options
/// \attention Do not use this directly, use class Option.
class BaseOption
  {
  public:
    std::string Description;  ///< Describes the purpose of this option
    std::string Command;      ///< 
    bool Strictly_needed;     ///< Defines if optional or not
    bool Provided;            ///< true if it was provided by user.
    int Option_type;          ///< records the type of option

  BaseOption(std::string arg, std::string desc=std::string(), bool nd=OptionMap::optional, int optp=1)
    : Description(desc), Command(arg), Strictly_needed(nd), Provided(false), Option_type(optp)
    {
    }

  inline std::string description(void) const {return Description;}

  inline bool provided(void) const {return Provided;}

  template<class _Tp>
  inline std::ostream& show_option(std::ostream &out, const Option<_Tp>& op) const
    {
    out << std::right << std::setw(OptionMap::Max_arg_length)  << Command << " = ";
    
    std::ostringstream st;
    op.show_internal(st);
    if (Provided) st << std::string(" (provided");
    else          st << std::string(" (default");
    if (Strictly_needed) st << std::string("  and necessary)");
    else          st << std::string(" and optional)");

    out << std::left << std::setw(25) << st.str() << std::endl;
    out << std::right << std::setw(OptionMap::Max_arg_length+3) << " " << std::left << std::setw(OptionMap::Max_desc_length) << Description;
    out << std::left << std::setw(OptionMap::Max_type_desc_length) << 
        std::string(" [") + OptionMap::type_description<typename Option<_Tp>::internal_type>() + std::string("]") << " ";
    return out;
    }
  };

template<class _Tp>
/// \brief Class for defining command-line options.
class Option: public BaseOption
  {
  public:
  typedef _Tp internal_type;
  
  internal_type Value;

 Option(std::string arg, std::string desc=std::string(), bool nd=OptionMap::optional,internal_type val=internal_type())
    : BaseOption(arg,desc,nd,OptionMap::type_string<internal_type>()), Value(val)
  
    {
    OptionMap::add(arg,this);
    }

  inline operator internal_type() const {return Value;}

  inline void show_internal(std::ostringstream& st) const
    {
    if ((internal_type)(*this) != internal_type()) st << (internal_type)(*this);
    else  st << std::string("void");
    }

  inline std::ostream& show(std::ostream &out) const
    {
    return show_option(out,*this);
    }
  };

};
#endif
