// $Id: Potentials.h 655 2011-11-24 13:50:42Z ge $
/// \file Potentials.h
/// \brief contains the basic Potentials function class
///
/// $Revision: 655 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_EXP_POTENTIALS_H
#define GBC_EXP_POTENTIALS_H "$Id: Potentials.h 655 2011-11-24 13:50:42Z ge $"
#include <valarray>
#include <fstream>
#include "ParameterMap.h"
namespace gbc_exp
  {
  template<class _Tp=double>
  /// \brief Base class for potentials
  ///
  /// This base class primarily handles the parameters that are used by all Potentials.
  class BasePotential
    {
    public:
      typedef _Tp value_type;
      typedef std::valarray<value_type> vector_type;
      typedef ParameterMap<value_type>  parameter_map_type;

    private:
      std::string Name;           ///< A name for the potential
      std::string Prefix;         ///< The prefix

    public:
      parameter_map_type ParMap;  ///< Map of parameters used by all potentials that inherit this class

      BasePotential(void): Name(), Prefix(), ParMap() {}

      virtual ~BasePotential() {}

      inline std::string name(void) const {return Name;} 

      inline std::string prefix(void) const {return Prefix;} 

      inline std::string complete_name(void) const 
        {
        if (Prefix != std::string())
          return Prefix+parameter_map_type::Field_separator+Name;
        else
          return Name;
        } 

      inline void name(std::string nm) {Name=nm;} 

      /// \brief Inserts a prefix in front of the name
      inline void prefix(std::string p) {Prefix=p;}


    /// \brief This is a placeholder. Each potential should define its own.
    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      return true;
      }

    /// \brief Retrieves the parameter from ParameterMap and places it into the local map.
    inline bool define_parameter(parameter_map_type& pm, std::string field, value_type& par)
      {
      pm.get_parameter(complete_name()+"."+field,par); //Get the parameter and places it value into par
      ParMap[complete_name()+"."+field]=par; //Now set the local map
      return true;
      }

    friend std::ostream& operator<<(std::ostream &out, const BasePotential& pot)
      {
      typename parameter_map_type::const_iterator it;
      out << "[" << pot.complete_name() << "]" << std::endl; 
      for(it=pot.ParMap.begin(); it != pot.ParMap.end(); ++it)
          out << it->first << " " << it->second << std::endl;
      return out;
      }

    inline virtual value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      std::cerr << "Warning: evaluate_xy() is only a placeholeder in file " 
                << __FILE__ << " at line " << __LINE__ << std::endl;
      return value_type();
      }

    inline vector_type evaluate(const vector_type& x) const
      {
      size_t x_size=x.size();
      vector_type result(x_size);
      for (size_t i=0; i < x_size; i++) result[i]=evaluate_xy(x[i]); 
      return result;
      }
      
    inline vector_type evaluate(const vector_type& x, const value_type& y) const
      {
      size_t x_size=x.size();
      vector_type result(x_size);
      for (size_t i=0; i < x_size; i++) result[i]=evaluate_xy(x[i],y); 
      return result;
      }

    inline vector_type evaluate(const vector_type& x, const vector_type& y) const
      {
      size_t nx=x.size(), ny=y.size();
      vector_type result(nx*ny);
      for (size_t i=0; i < nx; i++) 
        {
        for (size_t j=0; j < ny; j++) result[i*ny+j]=evaluate_xy(x[i],y[j]); 
        }
      return result;
      }

    };

  template<class _Tp>
  /// \brief Symmetrizes a one dimensional potential
  ///
  /// The potential \f$ V(x) \f$ is transformed
  /// into a symmetric potential depending on two variables \f$ (x,y) \f$,
  /// \f$ V(x,y)=\frac{1}{2}\left[V(x)+V(y)\right] \f$
  class SymmPotential: public BasePotential<typename _Tp::value_type>
    {
    public:
      typedef _Tp potential_type;
      typedef typename potential_type::value_type value_type;
      typedef std::valarray<value_type> vector_type;
      typedef BasePotential<value_type> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      
      potential_type Vx; ///< The \f$x\f$-dependend potential
      potential_type Vy; ///< The \f$y\f$-dependend potential

    SymmPotential(void): Vx(), Vy() {}

    template<class _Arg>
    SymmPotential(const _Arg& pref): Vx(pref), Vy(pref) {}

    template<class _Arg1, class _Arg2>
    SymmPotential(const _Arg1& prefVx,const _Arg2& prefVy): Vx(prefVx), Vy(prefVy) {}

    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      Vx.get_parameters(pm);
      Vy.get_parameters(pm);
      return true;
      }

    inline bool parameters(const potential_type& px) 
      {
      bool modified_vx=Vx.parameters(px);
      bool modified_vy=Vy.parameters(px);
      return modified_vx || modified_vy;
      }

    inline bool parameters(const potential_type& px, const potential_type& py) 
      {
      bool modified_vx=Vx.parameters(px);
      bool modified_vy=Vy.parameters(py);
      return modified_vx || modified_vy;
      }
 
    inline bool parameters(const SymmPotential& p) 
      {
      bool modified_vx=Vx.parameters(p.Vx);
      bool modified_vy=Vy.parameters(p.Vy);
      return modified_vx || modified_vy;
      }

    friend std::ostream& operator<<(std::ostream &out, const SymmPotential& sp)
      {
      out << sp.Vx << sp.Vy;
      return out;
      }


     /// \brief Evaluates the symmetrized potential 
     ///
     /// \f$V(x,y)=\frac{1}{2}\left[V(x)+V(y)\right] \f$
    inline value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      return 0.5*(Vx.evaluate_xy(x)+Vy.evaluate_xy(y));
      }
    };

  template<class _FirstPot,class _SecondPot>
  /// \brief Combines two potentials.
  ///
  /// The two potentials can be individually accessed through the First/Second
  /// objects. 
  class CmbPotential: public BasePotential<typename _FirstPot::value_type>
    {
    public:
      typedef _FirstPot first_potential_type;
      typedef _SecondPot second_potential_type;
      typedef typename first_potential_type::value_type value_type;
      typedef std::valarray<value_type> vector_type;
      typedef BasePotential<value_type> base_type;
      typedef typename base_type::parameter_map_type parameter_map_type;
      
      first_potential_type First;
      second_potential_type Second;

    CmbPotential(void): First(), Second() {}

    template<class _Arg>
    CmbPotential(const _Arg& arg): First(arg), Second(arg)
      {
      };

    template<class _Arg1, class _Arg2>
    CmbPotential(const _Arg1& p1, const _Arg2& p2): First(p1), Second(p2)
      {
      };

    template<class _Arg1, class _Arg2, class _Arg3>
    CmbPotential(const _Arg1& p1, const _Arg2& p2, const _Arg3& p3)
      : First(p1,p2), Second(p3)
      {
      };


    inline virtual bool get_parameters(parameter_map_type& pm)
      {
      First.get_parameters(pm);
      Second.get_parameters(pm);
      return true;
      }

    inline bool parameters(const first_potential_type& pf, const second_potential_type& ps) 
      {
      bool modified_f=First.parameters(pf);
      bool modified_s=Second.parameters(ps);
      return modified_f || modified_s;}

    inline bool parameters(const CmbPotential& p) 
      {return parameters(p.First,p.Second);}

    friend std::ostream& operator<<(std::ostream &out, const CmbPotential& cp)
      {
      out << cp.First << cp.Second;
      return out;
      }

    /// \brief Evaluates the combined potential by summing each potential.
    inline value_type evaluate_xy(value_type x, value_type y=value_type()) const
      {
      return First.evaluate_xy(x,y)+Second.evaluate_xy(x,y);
      }
    };

  };
#endif
