// $Id: SequenceDataset.h 698 2012-11-13 16:24:16Z ge $
/// \file SequenceDataset.h
/// \brief This class manages datasest of experimental sequence information
///
/// $Revision: 698 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_EXP_SEQUENCEDATASET_H
#define GBC_EXP_SEQUENCEDATASET_H "$Id: SequenceDataset.h 698 2012-11-13 16:24:16Z ge $"

#include "SequenceInfo.h"
#include <deque>
#include <vector>
#include "MathAux.h"

namespace gbc_exp
{
template<class _Tp=double>
class DataStatistics
  {
  typedef _Tp value_type;
  public:
    std::string name;
    value_type error;
    value_type average;
    value_type deviation;
    value_type std_deviation;
    value_type sqr_diff;

  DataStatistics(void)
    : name(), error(), average(), deviation(), 
      std_deviation(), sqr_diff() {}

  DataStatistics(std::string nm)
    : name(nm), error(), average(), deviation(), 
      std_deviation(), sqr_diff() {}

  inline void print_name(std::ostream &out) const
    {
    out << name << ".error " 
        << name << ".average "
        << name << ".std_deviation "
        << name << ".sqr_diff"; 
    }

  inline friend std::ostream& operator<<(std::ostream &out, DataStatistics& dt)
    {
    out << dt.error << " " << dt.average << " " << dt.std_deviation << " " << dt.sqr_diff;
    return out;
    }
  };

template<class _SequenceInfo>
class SequenceDataset
  {
  public:
    typedef _SequenceInfo                        sequence_info_type;
    typedef typename sequence_info_type::sequence_type
                                                 sequence_type;
    typedef size_t                               sequence_length_type;
    typedef typename sequence_info_type::value_type   
                                                 value_type;
    typedef value_type                           salt_concentration_type;
    typedef std::set<salt_concentration_type>    salt_set_type;
    typedef std::set<value_type>                 length_set_type;
    typedef std::deque<sequence_info_type>       sequence_info_deque_type;
    typedef std::deque<sequence_info_type*>      sequence_info_deque_ptype;
    typedef std::map<sequence_length_type,
                     sequence_info_deque_ptype>  length_dataset_type;
    typedef std::map<salt_concentration_type,
                     length_dataset_type>        salt_length_dataset_type;
    typedef typename salt_length_dataset_type::iterator 
                                                 iterator;
    typedef std::valarray<value_type>            vector_type;
    typedef DataStatistics<value_type>           datastatistics_type;

  std::string Dataset_identfifier;
  std::string Dataset_type;
  bool Uniform_species_concentration;
  value_type Species_concentration;
  bool Uniform_salt_concentration;
  value_type Salt_concentration;
  bool Adjust_temperatures;                               ///< Flag to check if we should adjust temperatures of sequences which are self-complimentary
  std::string Dataset_filename;
  sequence_info_deque_type  Raw_dataset;
  sequence_info_deque_type  Original_raw_dataset;
  salt_set_type             Salt_concentrations;          ///< Set of salt concentrations
  length_set_type           Lengths;                      ///< Set of lenght values

  length_dataset_type       Length_dataset;               ///< A map indexed by length
  salt_length_dataset_type  Salt_length_dataset;          ///< a map index by salt and length
  salt_length_dataset_type  Original_salt_length_dataset;

  datastatistics_type Temperature;
  datastatistics_type Entropy;
  datastatistics_type Enthalpy;

  std::deque<value_type> diff;

  int prediction_method;

  SequenceDataset(const std::string id=std::string()): Dataset_identfifier(id), Dataset_type(),
    Uniform_species_concentration(true), Species_concentration(),
    Uniform_salt_concentration(true), Salt_concentration(), Adjust_temperatures(false) {};
  
   void make_salt_length_map(void)
      {
      Salt_length_dataset.clear();
      typename sequence_info_deque_type::iterator si=Raw_dataset.begin();
      for(si=Raw_dataset.begin(); si != Raw_dataset.end(); si++)
        { 
        if (Salt_length_dataset.find(si->salt_concentration) == Salt_length_dataset.end())
          {
          Salt_concentrations.insert(si->salt_concentration);
          Salt_length_dataset.insert(typename salt_length_dataset_type::value_type(si->salt_concentration,length_dataset_type()));
          }

        if (Salt_length_dataset[si->salt_concentration].find(si->length) == Salt_length_dataset[si->salt_concentration].end())
          {
          Lengths.insert(si->length);
          Salt_length_dataset[si->salt_concentration].insert(typename length_dataset_type::value_type(si->length,sequence_info_deque_ptype()));
          }
      
        // Populate the map.
        Salt_length_dataset[si->salt_concentration][si->length].push_back(&(*si));
        }
      }

    sequence_info_type add(const std::string& seq,        ///< The sequence string 
             const std::string& comp,       ///< The complimentary sequence string 
             const value_type& salt_conc,   ///< The salt concentration
             const value_type& temperature, ///< The experimental melting temperature
             const value_type& enthalpy,    ///< The experimental enthalpy
             const value_type& entropy,     ///< The experimental entropy
             const value_type& spec_conc)   ///< Species concentration.
      {
      // If there is no reference species concentration yet, then use the current one.
      if (Species_concentration == value_type()) Species_concentration=spec_conc;
      // Check if species concentration is uniform for dataset
      if (Uniform_species_concentration)
        {
        if (spec_conc != Species_concentration) {Uniform_species_concentration=false;}
        }

      // If there is no reference salt concentration yet, then use the current one.
      if (Salt_concentration == value_type()) Salt_concentration=salt_conc;
      // Check if species concentration is uniform for dataset
      if (Uniform_salt_concentration)
        {
        if (salt_conc != Salt_concentration) {Uniform_salt_concentration=false;}
        }

      sequence_info_type si;
      si.length=seq.length();

      if (spec_conc == value_type()) si.species_concentration=Species_concentration;
      else                           si.species_concentration=spec_conc;

      si.salt_concentration=salt_conc;
      si.sequence=sequence_type(seq,comp); //this will become obsolete
      si.insert_sequence(seq,comp);
      si.determine_alpha();
      si.temperature.measured=temperature;
      si.enthalpy.measured=enthalpy;
      si.entropy.measured=entropy;
      si.Ref_concentration=Species_concentration;

      Raw_dataset.push_back(si);
      return si;
      }
      
    /// Update
    void update(sequence_info_type &si)
      {
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
	if (si.sequence == sequence_info_it->sequence)
	  {
	  *sequence_info_it = si;
	  }
	}
      }


    /// \brief Adds a new sequence and its experimental data to a map.
    ///
    /// This is one of the main functions of this class. It adds a new sequence
    /// and its experimental melting temperature, together with salt and species
    /// concentration to a map. The equivalence index is also calculated at
    /// this stage. The melting temperature is first corrected to the reference
    /// species concentration. If this is the first sequence which is being
    /// added, then it's species concentration is taken as reference and all
    /// further additions are corrected to this concentration.
    sequence_info_type add_temperature(const std::string& seq,        ///< The sequence string 
             const std::string& comp,       ///< The complimentary sequence string 
             const value_type& salt_conc,   ///< The salt concentration
             const value_type& temperature, ///< The experimental melting temperature
             const value_type& spec_conc)   ///< Species concentration.
      {
      add(seq,comp,salt_conc,temperature,0.0,0.0,spec_conc);
      sequence_info_type *si = &(Raw_dataset.back());
      if (Adjust_temperatures) si->adjust_temperature(); 
      else                     si->temperature.adjusted=si->temperature.measured;
      return *si;
      }

    /// \brief Adds a new sequence and its experimental data to a map.
    ///
    /// This is one of the main functions of this class. It adds a new sequence
    /// and its experimental melting temperature, together with salt and species
    /// concentration to a map. The equivalence index is also calculated at
    /// this stage. The melting temperature is first corrected to the reference
    /// species concentration. If this is the first sequence which is being
    /// added, then it's species concentration is taken as reference and all
    /// further additions are corrected to this concentration.
    sequence_info_type add_gibbs(const std::string& seq,        ///< The sequence string 
             const std::string& comp,       ///< The complimentary sequence string 
             const value_type& salt_conc,   ///< The salt concentration
             const value_type& enthalpy,    ///< The experimental enthalpy
             const value_type& entropy,     ///< The experimental entropy
             const value_type& spec_conc)   ///< Species concentration.
      {
      add(seq,comp,salt_conc,0.0,enthalpy,entropy,spec_conc);
      sequence_info_type *si = &(Raw_dataset.back());
      si->determine_temperature_from_gibbs();
      return *si;
      }

    void read_xml(std::string xmlfile, unsigned int restrict_length, std::ostream& out)
      {
      std::ifstream xml;
      xml.open(xmlfile.c_str());
      std::string type;
      xml >> type;
      std::string seq, comp;
      double salt, concentration;
      if (type == std::string("gibbs"))
        {
        Dataset_type="gibbs";
        double enthalpy, entropy;
        while(!xml.eof())
          {
          xml >> seq >> comp 
              >> enthalpy        //in kcal/mol
              >> entropy         //in cal/(mol.K) or eu
              >> salt            //in mM or mmol/L
              >> concentration ; //\mu M
          if (!xml.eof())
            {
            add_gibbs(seq,comp,salt,enthalpy,entropy,concentration);
            }
          }
        }      
      else
        {
        Dataset_type="temperature";
        double temp;
        while(!xml.eof())
          {
          xml >> seq >> comp 
              >> temp 
              >> salt            //in C
              >> concentration; //\mu M
          if (!xml.eof())
            {
            add_temperature(seq,comp,salt,temp,concentration);
            }
          }
        }
      make_salt_length_map();
      Original_raw_dataset=Raw_dataset;
      }

    void randomise_dataset(void)
      {
      Raw_dataset=Original_raw_dataset;
      make_salt_length_map();

      std::vector<value_type> temperature_change,enthalpy_change,entropy_change;
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
         if (Dataset_type == std::string("temperature"))
           {
           temperature_change.push_back(gauss_ran(Temperature.error));
           sequence_info_it->temperature.measured += temperature_change.back();
           if (Adjust_temperatures) sequence_info_it->adjust_temperature();
           else                     sequence_info_it->temperature.adjusted=sequence_info_it->temperature.measured;
           }
         if (Dataset_type == std::string("gibbs"))
           {
           enthalpy_change.push_back(gauss_ran(sequence_info_it->enthalpy.measured*Enthalpy.error*0.01));
           entropy_change.push_back(gauss_ran(sequence_info_it->entropy.measured*Entropy.error*0.01));
           sequence_info_it->enthalpy.measured += enthalpy_change.back();
           sequence_info_it->entropy.measured += entropy_change.back() ;
           value_type tmpch=sequence_info_it->temperature.measured;
           sequence_info_it->determine_temperature_from_gibbs();
           temperature_change.push_back(tmpch-sequence_info_it->temperature.measured);
           }
         }
      Temperature.std_deviation=standard_deviation(temperature_change);
      Enthalpy.std_deviation=standard_deviation(enthalpy_change);
      Entropy.std_deviation=standard_deviation(entropy_change);
      }



    /// \brief Calculates the difference between the predicted and the adjusted temperature
    inline void difference(void)
      {
      diff.clear();
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
        diff.push_back(fabs(sequence_info_it->temperature.predicted-sequence_info_it->temperature.adjusted));
        }
      Temperature.average=average(diff);
      Temperature.deviation=standard_deviation(diff);
      Temperature.sqr_diff=sqr_average(diff);
      }

    /// \brief Prints out the experimental and calculated temperatures.
    inline void print_temperatures(std::ostream& out)
      {
      typename sequence_info_deque_type::iterator sequence_info_it=Raw_dataset.begin();
      sequence_info_it->print_head(out);
      for(sequence_info_it=Raw_dataset.begin(); sequence_info_it != Raw_dataset.end(); sequence_info_it++)
        {
        sequence_info_it->print(out);
        }
      out << "StdTempDiff "     << Temperature.std_deviation << std::endl;
      out << "StdEnthalpyDiff " << Enthalpy.std_deviation    << std::endl;
      out << "StdEntropyDiff "  << Entropy.std_deviation     << std::endl;

      }

   inline void verify(std::ostream& out)
      {
      difference();
      out << Temperature.average << " " << Temperature.deviation << " " << Temperature.sqr_diff;
      }

   inline void print_tex(std::ostream& out)
     {
     out << "\\SqrDiff{" << Temperature.sqr_diff   << "}" << std::endl;
     out << "\\AvDiff{"  << Temperature.average   << "}" << std::endl;
     out << "\\StdDiff{" << Temperature.deviation << "}" << std::endl;
     out << "\\StdTempDiff{" << Temperature.std_deviation << "}" << std::endl;
     out << "\\StdEnthalpyDiff{" << Enthalpy.std_deviation << "}" << std::endl;
     out << "\\StdEntropyDiff{" << Entropy.std_deviation << "}" << std::endl;
     }

    /// \brief Build a vector of lengths for a given salt concentration
    inline vector_type length_vector(typename salt_length_dataset_type::iterator& salt_it)
      {
      vector_type lv(salt_it->second.size());
      typename length_dataset_type::iterator length_it;
      size_t n;
      for (length_it=salt_it->second.begin(),n=0; length_it != salt_it->second.end(); ++length_it, ++n)
        {
        lv[n]=length_it->first;
        }
      return lv;
      }

    /// \brief Build a vector of thermal_equivalences for a given length
    inline vector_type thermal_equivalence_vector(typename length_dataset_type::iterator& length_it)
      {
      vector_type lv(length_it->second.size());
      typename sequence_info_deque_ptype::iterator si_it;
      size_t n;
      for (si_it=length_it->second.begin(),n=0; si_it != length_it->second.end(); ++si_it, ++n)
        {
        lv[n]=(*si_it)->thermal_equivalence;
        //std::cerr << "theq=" << lv[n] << std::endl;
        }
      return lv;
      }

    /// \brief Build a vector of all thermal_equivalences
    inline vector_type thermal_equivalence_vector(void)
      {
      vector_type lv(Raw_dataset.size());
      typename sequence_info_deque_type::iterator si=Raw_dataset.begin();
      size_t n;
      for(si=Raw_dataset.begin(),n=0; si != Raw_dataset.end(); si++, ++n)
        {
	//std::cerr << "theq=" << si->thermal_equivalence << std::endl;
        lv[n]=si->thermal_equivalence;
        }
      return lv;
      }

   /// \brief Build a vector of temperatures for a given length
   inline  vector_type temperature_vector(typename length_dataset_type::iterator& length_it)
      {
      vector_type lv(length_it->second.size());
      typename sequence_info_deque_ptype::iterator si_it;
      size_t n;
      for (si_it=length_it->second.begin(),n=0; si_it != length_it->second.end(); ++si_it, ++n)
        {
        lv[n]=(*si_it)->temperature.adjusted;
        //std::cerr << "temp=" << lv[n] << std::endl;
        }
      return lv;
      }

   /// \brief Build a vector of all temperatures
   inline  vector_type temperature_vector(void)
      {
      vector_type lv(Raw_dataset.size());
      typename sequence_info_deque_type::iterator si=Raw_dataset.begin();
      size_t n;
      for(si=Raw_dataset.begin(),n=0; si != Raw_dataset.end(); si++, ++n)
        {
        lv[n]=si->temperature.adjusted;
        //std::cerr << "temp=" << lv[n] << std::endl;
        }
      return lv;
      }
  };
};
#endif