// $Id: SequenceInfo.h 628 2011-08-31 15:42:16Z ge $
/// \file SequenceInfo.h
/// \brief This class manages experimental sequence information
///
/// $Revision: 628 $
/// \author Gerald Weber <gweberbh@gmail.com>
#ifndef GBC_EXP_SEQUENCEINFO_H
#define GBC_EXP_SEQUENCEINFO_H "$Id: SequenceInfo.h 628 2011-08-31 15:42:16Z ge $"

#include "NucleotideSequence.h"
#include "NeighbourSequence.h"
#include "Duplex.h"

using namespace gbc;
namespace gbc_exp
{
template<class _Tp=double>
class Data
  {
  typedef _Tp value_type;
  public:
    std::string name;
    value_type measured;
    value_type adjusted;
    value_type predicted;

  Data(void): name(), measured(), adjusted(), predicted() {}

  Data(std::string nm): name(nm), measured(), adjusted(), predicted() {}

  inline void print_name(std::ostream &out) const
    {
    out << name << ".measured " 
        << name << ".adjusted "
        << name << ".predicted "; 
    }

  inline friend std::ostream& operator<<(std::ostream &out, Data& dt)
    {
    out << dt.measured << " " << dt.adjusted << " " << dt.predicted;
    return out;
    }
  };

template<class _Tp=double, class _sequence_type=std::pair<std::string,std::string> >
class SequenceInfo
  {
  public:
    typedef _Tp                 value_type;
    typedef Data<value_type>    data_type;
    typedef _sequence_type      sequence_type;
    typedef Duplex<>            duplex_type;
    typedef NeighbourSequence<> neighboursequence_type;

    sequence_type sequence;
    static const value_type Concentration_factor=1e-6; ///< All species concentrations are in milimol
    static const value_type R=1.987;   ///< Gas constant, cal/(K mol)
    value_type Ref_concentration;     ///< Reference species concentration
    value_type alpha;                 ///< Determines if sequence is selfcompementary
    value_type species_concentration; ///< Oligomer species concentration in mM
    value_type salt_concentration;    ///< Salt concentration in mM
    data_type  temperature;           ///< Sequence temperature to species concentration
    data_type  enthalpy;              ///< Sequence Enthalpy as given in papers
    data_type  entropy;               ///< Sequence Entropy as given in papers
    int length;                       ///< Sequence length
    int prediction_method;            ///< Flag for selecting prediction methods
    value_type thermal_equivalence;   ///< Thermal equivalence
    value_type thermal_index;         ///< Thermal index sqrt(thermal_equivalence)

    /// Sequence information
    duplex_type            Exact_duplex;        ///< Contains the sequence in Duplex format, it is exact because we are not applying symmetry reductions
    duplex_type            Reduced_duplex;      ///< Contains the symmetry reduced sequence in Duplex format
    neighboursequence_type Reduced_neighbours;  ///< Contains the symmetry reduced sequence in Nearest neighbour format
    /// Sequence analysis
    bool                Has_CG;                 ///< True if sequence has at least one CG
    bool                Is_AT_only;             ///< True if sequence is AT only
    bool                Self_complementary;     ///< True if sequence is self-complementary
    int                 AT_five_prime_terminal; ///< Number of 5'TA3' terminals, 
 
    SequenceInfo(void): sequence(), alpha(), species_concentration(), 
                        salt_concentration(), 
                        temperature("temperature"), enthalpy("enthalpy"), entropy("entropy"),
                        prediction_method(), thermal_equivalence(), thermal_index(),
                        Has_CG(false), Is_AT_only(false), 
                        Self_complementary(false), AT_five_prime_terminal()  {}

    inline void print_head(std::ostream& out)
      {
      out << "Main/Complementary " 
          << "alpha "
          << "salt_concentration "
          << "species_concentration ";
      temperature.print_name(out);
      enthalpy.print_name(out);
      entropy.print_name(out);
      out << "prediction_method " 
          << "thermal_equivalence "
          << "thermal_index "
          << std::endl;
      }

    void print(std::ostream& out)
      {
      out << sequence.first << "/" << sequence.second 
          << " " << alpha
          << " " << salt_concentration
          << " " << species_concentration
          << " " << temperature  
          << " " << enthalpy 
          << " " << entropy 
          << " " << prediction_method 
          << " " << thermal_equivalence
          << " " << thermal_index
          << std::endl;
      }

    inline void insert_sequence(const std::string& seq, const std::string& comp=std::string())
      {
      Exact_duplex.clear();
      Reduced_duplex.clear();
      if(comp != std::string())
        {
        Exact_duplex=Duplex<>(seq+"/"+comp,true,BasePair<>::context_dependent);
        Reduced_duplex=Duplex<>(seq+"/"+comp,true,BasePair<>::not_context_dependent);
        }
      else 
        {
        Exact_duplex=Duplex<>(seq,true,BasePair<>::context_dependent);
        Reduced_duplex=Duplex<>(seq,true,BasePair<>::not_context_dependent);
        }

      Reduced_neighbours.clear();
      Reduced_neighbours.Periodic=false; // This avoids that the first and last base-pair form a nearest-neighbour pair

      Reduced_neighbours=Exact_duplex;
      Reduced_neighbours.reduce_to_smallest_symmetry();

      Self_complementary=Exact_duplex.self_complementary();

      BasePair<> AT("AT",BasePair<>::context_dependent),
                 TA("TA",BasePair<>::context_dependent),
                 CG("CG",BasePair<>::not_context_dependent);
      typename duplex_type::iterator it;
      AT_five_prime_terminal=0;
      //This accounts for a 5'T.A3' at the beginning
      if ((*Exact_duplex.begin()) == TA) AT_five_prime_terminal++;
      //This accounts for a 3'A.T5' at the end
      if ((*Exact_duplex.rbegin())  == AT) AT_five_prime_terminal++;

      Has_CG=false;
      Is_AT_only=true;
      for (it=Reduced_duplex.begin(); it != Reduced_duplex.end(); ++it)
        {
        if (*it == CG) Has_CG=true;
        if (not (*it == AT)) Is_AT_only=false; 
        }
      }

    /// \brief Checks for selfcomplementarity and sets alpha value
    ///
    /// \returns 1 if self-complementary, 4 otherwise
    inline value_type determine_alpha(void)
      {
      NucleotideSequence<> s(sequence.first.length()),c(sequence.second.length());
      s << sequence.first; c << sequence.second; c.reverse();
      if ((std::string)s == (std::string)c) alpha=1.0;
      else                                  alpha=4.0;
      return alpha;
      }

    /// \brief Adjuts temperature to current reference concentration
    ///
    /// \attention This temperature adjustment is too approximate and should not be used.
    inline void adjust_temperature(void)
      {
      if (Ref_concentration != species_concentration)
        {
        // Note: it is correct to use the same alpha for conc1 and conc2
        temperature.adjusted=temperature.measured*log(species_concentration*Concentration_factor/alpha)/log(Ref_concentration*Concentration_factor/alpha);
        }
      else temperature.adjusted=temperature.measured;
      }

    inline void determine_temperature_from_gibbs(void)
      {
      temperature.measured=enthalpy.measured*1000.0/(entropy.measured+R*log(species_concentration*Concentration_factor/alpha))-273.15;
      temperature.adjusted=temperature.measured;
      }

   };
};
#endif
