// $Id: TransferMatrix.h 702 2013-02-06 16:19:28Z ge $
/// \file TransferMatrix.h
/// \brief contains the Hamiltonian function class
///
/// $Revision: 702 $
/// \author Gerald Weber <gweber@gmail.com>
#ifndef GBC_EXP_TRANSFERMATRIX_H
#define GBC_EXP_TRANSFERMATRIX_H "$Id: TransferMatrix.h 702 2013-02-06 16:19:28Z ge $"
#include <valarray>
#include <map>
#include "Range.h"
#include "Hamiltonian.h"
#include "Quadrature.h"
#include "EigenSystem.h"
#include <gsl/gsl_const_mksa.h>
#include <boost/filesystem.hpp>


namespace gbc_exp
  {
  template<class _Tp=double>
  /// \brief Calculation of Hamiltonians using the Transfer Matrix technique.
  ///
  /// Add new hamiltonians as objects (void constructor), add an short identification
  /// to hamiltonian_id and ammend the method select_hamiltonian and K.
  /// Also provide a specific parameter method for the new Hamiltonian.
  class TransferMatrix
    {
    public:
      typedef _Tp                              value_type;
      typedef std::valarray<value_type>        vector_type;
      typedef valmatrix<value_type>            matrix_type;
      typedef std::map<value_type,vector_type>  result_map_type;
      typedef Quadrature<Legendre<std::valarray<value_type> > > quadrature_type;

      enum hamiltonian_id {pb,dpb,hms,jb};  ///< The short identification for Hamiltonians

      bool Debug;                           ///< Flag for debugging
      bool Periodic_boundary_condition;     ///< Flag for boundary conditions
      bool Recalculate_diagonalization;     ///< If true recalculates the diagonalization
      size_t Eigenvalue_cutoff;             ///< Cut-off value for the number of Eigenvalues/vectors to be used in certain integration procedures.

    protected:
      value_type Beta;                      ///< The temperature divided by the Boltzman constant
      value_type Temperature;               ///< The temperature to be used for the hamiltonian
    public:
      value_type Lower_limit;               ///< Lower integration limit
      value_type Upper_limit;               ///< Upper integration limit
      vector_type Integration_space;        ///< Vector containg the points for integration, these are usually rescaled by the quadrature class
      vector_type Weights;                  ///< The integration calculated weights, depends on the selected quadrature.
      vector_type SqrtWeights;              ///< The square root of Weights, needed for symmetrizing the matrix. 
      size_t Integration_steps;             ///< The number of points used in the integration.
      quadrature_type* quadrature;          ///< The quadrature used for integration and formation of the eigensystem
      EigenSystem<value_type,value_type> eigen_engine; ///< The functions that diagonalizses the eigensystem
      
      // Hamiltoninans
      hamiltonian_id Hamiltonian_ID;        ///< The identification of the Hamiltonian to be used/
      PeyrardBishop<value_type>        PB;
      DauxoisPeyrardBishop<value_type> DPB;
      HarmonicMorseSolvent<value_type> HMS;
      JoyeuxBuyukdagli<value_type>     JB;

      value_type Position;
      std::string      ParameterIdName;
      std::string      MatrixDir;


      TransferMatrix(void)
       : Debug(false), Periodic_boundary_condition(true), Recalculate_diagonalization(true),
         Eigenvalue_cutoff(), Temperature(), Lower_limit(), Upper_limit(), Integration_steps(),
         Hamiltonian_ID(pb)
       {
       eigen_engine.Decrease_sort_eigenvalues=true;
       quadrature = new quadrature_type();
       quadrature->Rule.Save_calculations=true;
       quadrature->Rule.Load_calculations=true;
       }

      TransferMatrix(const BasePairNeighbours<> &bpn)
       : Debug(false), Periodic_boundary_condition(true), Recalculate_diagonalization(true),
         Eigenvalue_cutoff(), Temperature(), Lower_limit(), Upper_limit(), Integration_steps(0),
         Hamiltonian_ID(pb), PB(bpn), DPB(bpn), HMS(bpn), JB(bpn)
       {
       eigen_engine.Decrease_sort_eigenvalues=true;
       quadrature = new quadrature_type();
       quadrature->Rule.Save_calculations=true;
       quadrature->Rule.Load_calculations=true;
       }

    void eigenvalue_cutoff(size_t c) {Eigenvalue_cutoff=c;}

    /// \brief Selects the Hamiltonian to be used in the calculation.
    ///
    /// \attention This method needs to be modified if new Hamiltonians are added.
    void select_hamiltonian(std::string hamiltonian_name)
      {
      hamiltonian_id new_id=hamiltonian_id();
      if (hamiltonian_name==std::string("pb"))  new_id=pb;
      if (hamiltonian_name==std::string("dpb")) new_id=dpb;
      if (hamiltonian_name==std::string("hms")) new_id=hms;
      if (hamiltonian_name==std::string("jb"))  new_id=jb;
      if (new_id != Hamiltonian_ID) 
        {
        Recalculate_diagonalization=true;
        Hamiltonian_ID=new_id;
        }
      }

    inline std::string hamiltonian(void)
      {
      if (Hamiltonian_ID==pb)  return std::string("pb");
      if (Hamiltonian_ID==dpb) return std::string("dpb");
      if (Hamiltonian_ID==hms) return std::string("hms");
      if (Hamiltonian_ID==jb)  return std::string("jb");
      return std::string();
      }

    inline bool get_parameters(ParameterMap<value_type>& pm)
      {
      ParameterIdName=pm.Identify;
      if (Hamiltonian_ID==pb)  return PB.get_parameters(pm);
      if (Hamiltonian_ID==dpb) return DPB.get_parameters(pm);
      if (Hamiltonian_ID==hms) return HMS.get_parameters(pm);
      if (Hamiltonian_ID==jb)  return JB.get_parameters(pm);
      return false;
      }

    friend std::ostream& operator<<(std::ostream &out, const TransferMatrix& tm)
      {
      if (tm.Hamiltonian_ID==TransferMatrix::pb)  out << tm.PB;
      if (tm.Hamiltonian_ID==TransferMatrix::dpb) out << tm.DPB;
      if (tm.Hamiltonian_ID==TransferMatrix::hms) out << tm.HMS;
      if (tm.Hamiltonian_ID==TransferMatrix::jb)  out << tm.JB;
      return out;
      }


    /// \brief Assigns parameters for the Peyrard-Bishop Hamiltonian
    inline bool parameters(const PeyrardBishop<value_type>& p) 
      {
      bool modified=PB.parameters(p);
      Recalculate_diagonalization=Recalculate_diagonalization || modified;
      return modified;
      }

    /// \brief Assigns parameters for the Daxuois-Peyrard-Bishop Hamiltonian
    inline bool parameters(const DauxoisPeyrardBishop<value_type>& p) 
      {
      bool modified=DPB.parameters(p);
      Recalculate_diagonalization=Recalculate_diagonalization || modified;
      return modified;
      }

    /// \brief Assigns parameters for the harmonic Morse-Solvent Hamiltonian
    inline bool parameters(const HarmonicMorseSolvent<value_type>& p) 
      {
      bool modified=HMS.parameters(p);
      Recalculate_diagonalization=Recalculate_diagonalization || modified;
      return modified;
      }

    /// \brief Assigns parameters for the harmonic Morse-Solvent Hamiltonian
    inline bool parameters(const JoyeuxBuyukdagli<value_type>& p) 
      {
      bool modified=JB.parameters(p);
      Recalculate_diagonalization=Recalculate_diagonalization || modified;
      return modified;
      }

    /// \brief Evaluates the K function.
    ///
    /// \attention This method needs to be modified if new Hamiltonians are added.
    inline vector_type K(const vector_type& x, const vector_type& y)
      {
      if (Hamiltonian_ID==pb)  return exp(-Beta*PB.evaluate(x,y));
      if (Hamiltonian_ID==dpb) return exp(-Beta*DPB.evaluate(x,y));
      if (Hamiltonian_ID==hms) return exp(-Beta*HMS.evaluate(x,y));
      if (Hamiltonian_ID==jb)  return exp(-Beta*JB.evaluate(x,y));
      return vector_type();
      }

    /// \brief Evaluates the Y function.
    ///
    /// \attention This method needs to be modified if new Hamiltonians are added.
    inline vector_type Y(const vector_type& x)
      {
      if (Hamiltonian_ID==pb)  return exp(-0.5*Beta*PB.potential()->First.Vx.evaluate(x));
      if (Hamiltonian_ID==dpb) return exp(-0.5*Beta*DPB.potential()->First.Vx.evaluate(x));
      if (Hamiltonian_ID==hms) return exp(-0.5*Beta*HMS.potential()->First.Vx.evaluate(x));
      if (Hamiltonian_ID==jb)  return exp(-0.5*Beta*JB.potential()->First.Vx.evaluate(x));
      return vector_type();
      }

    /// \brief Evaluates the A function.
    ///
    /// \attention This method needs to be modified if new Hamiltonians are added.
    inline vector_type A(const vector_type& x, const vector_type& y)
      {
      if (Hamiltonian_ID==pb)  return exp(-Beta*PB.basepair_interaction()->evaluate(x,y));
      if (Hamiltonian_ID==dpb) return exp(-Beta*DPB.basepair_interaction()->evaluate(x,y));
      if (Hamiltonian_ID==hms) return exp(-Beta*HMS.basepair_interaction()->evaluate(x,y));
      if (Hamiltonian_ID==jb)  return exp(-Beta*JB.basepair_interaction()->evaluate(x,y));
      return vector_type();
      }

    /// \brief Sets the new temperature.
    ///
    /// Also recalculates the Beta value.
    void temperature(double T) 
      {
      if (T != Temperature) Recalculate_diagonalization=true;
      Temperature=T; 
      Beta=1.0/(GSL_CONST_MKSA_BOLTZMANN/GSL_CONST_MKSA_ELECTRON_VOLT*T);
      }

    void calculate_eigenvalue_only(bool ceo=true) {eigen_engine.Calculate_eigenvalue_only=ceo;}

    /// \brief Sets the limits for the integration
    ///
    /// This method recalculated the integration weights and rescales the integration
    /// space if necessary.
    void limits(value_type l, ///< Lower integration limit, usually negative 
                value_type u, ///< Upper integration limit
                size_t s)     ///< Number of integration points
      {
      bool recalculate=false;
      if (Integration_steps != s)
        {
        Integration_steps=s;
        Integration_space.resize(Integration_steps);
        Weights.resize(Integration_steps);
        recalculate=true;
        SqrtWeights.resize(Integration_steps);
        }
      if ((Lower_limit !=l) || (Upper_limit!=u))
        {
        Lower_limit=l; Upper_limit=u; 
        recalculate=true;
        }
      if (recalculate)
        {
        Integration_space=linear_space<vector_type>(Lower_limit,Upper_limit,Integration_steps);
        quadrature->adjust(Integration_space);
        Weights=quadrature->Rule.weights();
        SqrtWeights=sqrt(Weights);
        Recalculate_diagonalization=true;
        }
      }

    /// \brief Executes the calculation.
    void diagonalize(void)
      {
      if (Recalculate_diagonalization)
        {
        valmatrix<value_type> kxy(Integration_steps,Integration_steps);
        kxy=K(Integration_space,Integration_space);
        valmatrix<value_type> R(Integration_steps,Integration_steps);
        R.diagonal(SqrtWeights);

        kxy=R*kxy*R;

        eigen_engine.diagonalize(kxy);

        if (!eigen_engine.Calculate_eigenvalue_only) 
          {
          R.diagonal(pow(SqrtWeights,value_type(-1.0)));
          eigen_engine.Eigenvector=R*eigen_engine.Eigenvector;
          normalize_eigenvectors();
          save_matrix(eigen_engine.Eigenvector,"eigenvectors");
          }
        Recalculate_diagonalization=false;
        }
      save_matrix(eigen_engine.Eigenvalue,"eigenvalues");
      }

    /// \brief Calculates the matrix \f$ C_{n,m}\f$
    valmatrix<value_type> C_matrix(const valmatrix<value_type>& base)
      {
      size_t N;
      if (Eigenvalue_cutoff > (size_t)0) N=Eigenvalue_cutoff;
      else N=Integration_steps;
      valmatrix<value_type> kxy(Integration_steps,Integration_steps);
      valmatrix<value_type> W(Integration_steps,Integration_steps);
      valmatrix<value_type> phi(Integration_steps,N);

      phi=base.submatrix(0,0,Integration_steps,N);

      kxy=K(Integration_space,Integration_space);
      W.diagonal(Weights);

      return transpose(phi)*W*kxy*W*phi;
      };

    /// \brief Calculates the matrix \f$ C_{n,m}\f$
    valmatrix<value_type> A_matrix(const valmatrix<value_type>& base)
      {
      size_t N;
      if (Eigenvalue_cutoff > (size_t)0) N=Eigenvalue_cutoff;
      else N=Integration_steps;
      valmatrix<value_type> axy(Integration_steps,Integration_steps);
      valmatrix<value_type> W(Integration_steps,Integration_steps);
      valmatrix<value_type> phi(Integration_steps,N);

      phi=base.submatrix(0,0,Integration_steps,N);
      axy=A(Integration_space,Integration_space);
      W.diagonal(Weights);

      return transpose(phi)*W*axy*W*phi;
      };

    /// \brief Calculates the matrix \f$ Y_{n,m}\f$
    valmatrix<value_type> Y_matrix(const valmatrix<value_type>& base)
      {
      size_t N;
      if (Eigenvalue_cutoff > (size_t)0) N=Eigenvalue_cutoff;
      else N=Integration_steps;
      valmatrix<value_type> kxy(Integration_steps,Integration_steps);
      valmatrix<value_type> Y(Integration_steps,Integration_steps);
      valmatrix<value_type> W(Integration_steps,Integration_steps);
      valmatrix<value_type> phi(Integration_steps,N);

      phi=base.submatrix(0,0,Integration_steps,N);
      kxy=K(Integration_space,Integration_space);
      Y.diagonal(Integration_space);
      W.diagonal(Weights);

      return transpose(phi)*Y*W*phi;
      };



    /// \brief Performs the eigenvector normalization.
    void normalize_eigenvectors(void)
      {
      size_t l;
      for (l=0; l < Integration_steps; ++l)
        {
        vector_type eig=eigenvector(l);
        double norm=vector_type(pow(eig,value_type(2.0))*Weights).sum();
        eig /= sqrt(norm);
        eigen_engine.Eigenvector.column(l,eig);
        }
      }

    /// \brief Retrieves the n-th calculated eigenvalue. 
    inline value_type eigenvalue(const size_t& n) const
      {return eigen_engine.Eigenvalue[n];}

    /// \brief Retrieves the n-th calculated eigenvalue. 
    inline vector_type eigenvalues(void) const
      {return eigen_engine.Eigenvalue.Array;}

    /// \brief Retrieves the n-th calculated eigenvector. 
    inline vector_type eigenvector(const size_t& n) const
      {return eigen_engine.Eigenvector.column(n).Array;}

    /// \brief Retrieves the n-th calculated eigenfunction. 
    inline value_type eigenfunction(const value_type& x, const size_t& n)
      {
      vector_type xv(x,1);
      if (Debug) {std::cerr << vector_type(K(xv,Integration_space)).sum() << " " << std::endl;}
      vector_type kxy(1,Integration_steps);
      kxy=K(xv,Integration_space);
      return vector_type(Weights*kxy*eigenvector(n)).sum()/eigenvalue(n);
      }

    /// \brief Calculates the partition function \f$Z\f$ for \f$N\f$ base-pairs.
    ///
    /// \f$ Z=\sum_{n=1}^{\infty}\lambda_n^N \f$
    value_type partition_function(const unsigned int& N) const
      {
      vector_type z=pow(eigenvalues(),static_cast<value_type>(N));
      return z.sum();
      }

    /// \brief Calculates the coefficient \f$a_m\f$ needed for open boundary functions.
    ///
    /// \f$ a_m=\int \phi_m(y)e^{-\frac{1}{2}\beta V(y)}dy \f$
    double coef_am(const size_t& m)
      {
      return vector_type(eigenvector(m)*Y(Integration_space)*Weights).sum();
      }

    /// \brief Calculates \f$\lange n| y| m\angle$\f.
    value_type average_y(const size_t& n,const size_t& m) const
      {
      return vector_type(eigenvector(n)*Integration_space*eigenvector(m)*Weights).sum();
      }

    /// \brief Calculates \f$\lange y \angle$\f.
    ///
    /// \f$ \langle y \rangle 
    /// =\frac{1}{Z_y}\sum_{n=1}^{\infty}\lambda_n^N\langle n|y|n\rangle \f$
    value_type average_y(const unsigned int& N) ///< Number of base pairs (starts at 1)
      const
      {
      vector_type z=pow(eigenvalues(),static_cast<value_type>(N))/partition_function(N);
      size_t n, mx=Eigenvalue_cutoff;
      if (Eigenvalue_cutoff == 0) mx=eigenvalues().size();
      value_type y=0.0;
      for (n=0; n < mx; ++n)
         {
         y += z[n]*average_y(n,n);
         }
      return y;
      }

    /// \brief Calculates \f$ \langle | y_m | \rangle \f$ with open boundary conditions.
    /// \f$ \langle y_m \rangle=
    /// \frac{\sum_{l,n}a_la_n\Delta_l^{N-m}\Delta_n^{m-1}\langle n|y_m|l\rangle}{\sum_n\Delta_n^{N-1}a_n^2}
    /// \f$
    vector_type open_average_y(const unsigned int& N) ///< Number of base pairs (starts at 1)
      {
      size_t n, cutoff=Eigenvalue_cutoff;
      if (Eigenvalue_cutoff == 0) cutoff=eigenvalues().size();
      vector_type Delta(cutoff);
      Delta=vector_type(eigen_engine.Eigenvalue.Array[std::slice(0,cutoff,1)])/eigenvalue(1);
      vector_type a(cutoff);
      double s1=0.0;
      vector_type s2(N);
      for (n=0; n < cutoff; ++n) 
        {
        a[n]=coef_am(n);
        s1+=pow(Delta[n],static_cast<value_type>(N-1))*pow(a[n],2.0);
        }
      for (size_t m=1; m <= N; ++m)
        {
        s2[m-1]=0.0;
        for (n=0; n < cutoff; ++n)
          {
          double n_only=a[n]*pow(Delta[n],static_cast<value_type>(m-1))/s1;
          for (size_t l=0; l < cutoff; ++l)
            s2[m-1]+=n_only*a[l]*pow(Delta[l],static_cast<value_type>(N-m))*average_y(n,l);
          }
        }
      return s2;
      }

    std::string matrix_filename(std::string mat_name)
      {
      std::stringstream filename;
      filename << MatrixDir << "/" << ParameterIdName
               << "-" << Temperature << "-" << mat_name;
      return filename.str();
      }

    /// \brief Stores matrices into the map and saves into a directory.
    ///
    /// \attention If you don't want to store the matrix in a file
    /// leave MatrixDir empty.
    void save_matrix(const matrix_type& mat,
                     std::string mat_name)
      {
      if (!MatrixDir.empty())
        {
        boost::filesystem::create_directory(MatrixDir);
        std::ofstream outfile;
        std::string filename;
        filename=matrix_filename(mat_name);
        outfile.open(filename.c_str());
        outfile << *this << std::endl;
        outfile << mat;
        outfile.close();
        quadrature->Rule.Base_directory=MatrixDir;
        }
      }

   };

 };

#endif
