// $Id: tfreg.cpp 702 2013-02-06 16:19:28Z ge $
/// \file tfreg.cpp
/// \brief Calculates temperatures using the regression method
/// \author Gerald Weber <gweberbh@gmail.com>
/// $Revision: 702 $

#include "JobControl.h"
#include "Options.h"
#include "NeighbourSequence.h"
#include "Regression.h"
#include <unistd.h>
#include "HeterogenousTM.h"
//#include "Actions.h"
#include "FileParser.h"
#include "RegressionModel.h"
namespace gbc_exp
  {
  template<>
  std::valarray<HeterogenousTM<>::value_type> ActionAverageY::action(HeterogenousTM<>& tf,size_t N)
    {
    number_of_basepairs=tf.Equivalent_duplex.size();
    tf.retrieve_eigensystem();
    tf.retrieve_matrices();
    return tf.average_y();
    }
  };

using namespace gbc;
using namespace std;
namespace gbc_test=gbc_exp;
using namespace gbc_test;

typedef double value_type;

int main(int argc,char *argv[])
  {
  JobControl::status("running",true);
  JobControl::write_log("$Id: tfreg.cpp 702 2013-02-06 16:19:28Z ge $");
 
  Option<string>         output_name("o","output files basename",OptionMap::necessary);
  Option<string>         reg_name("reg","regression parameter file",OptionMap::optional);
  Option<string>         result("res","select which type of results",OptionMap::optional,"regression");
  Option<string>         duplextype("duplextype","type of duplex, e.g. DNA or RNA",OptionMap::optional,"DNA");
  Option<deque<string> > par_file("par","list of parameter file",OptionMap::necessary);
  Option<string>         expand("expand","expand on this base",OptionMap::optional,"CG_CG");
  Option<string>         matrixbase("matrix","location of the matrices",OptionMap::optional);
  Option<string>         datafile("data","parsed data file",OptionMap::optional);
  Option<string>         debug_file("debugfile","file to be debugged",OptionMap::optional);
  Option<string>         sequence("seq","sequence (5' to 3')",OptionMap::optional);
  Option<string>         csequence("cseq","complementary sequence (3' to 5')",OptionMap::optional);
  Option<int>            cutoff("cutoff","number of eigenvalues at which to cutoff a summation",OptionMap::optional,0);
  Option<bool>           periodic_bc("pbc","periodic boundary condition",OptionMap::optional,false);
  Option<bool>           debug("debug","select debuging",OptionMap::optional,false);
  Option<std::string>    model("m","model",OptionMap::optional,"pb");
  Option<double>         temperature("t","temperature in Kelvin",OptionMap::optional,370.0);
  Option<Range<double> > integration("int","integration range",OptionMap::optional);
  Option<double>         exp_err("ee","experimental error",OptionMap::optional,0.0);
  Option<int>            ran_seed("rs","random_seed",OptionMap::optional,1);
  Option<int>            min_length_regression("mlr","minimum duplex length to be included in regression",OptionMap::optional,0);
  Option<int>            prediction_method("pm","prediction method",OptionMap::optional,2);
  Option<double>         salt("salt","salt concentration (mM)", OptionMap::optional,69);

  OptionMap::arguments(argc,argv);
  OptionMap::finished();

  if (ran_seed.provided()) srand(ran_seed);

  if (duplextype.Value=="RNA")
    {
    Nucleotide<>::dictionary.add('U','U');
    Nucleotide<>::dictionary.rm('T');
    Nucleotide<>::dictionary.complementary_pair('A','U');
    //std::cout << Nucleotide<>::dictionary;
    }

  HeterogenousTM<value_type> HTM(expand);
  HTM.debug(debug.Value);
  HTM.verbose(debug.Value);
  HTM.matrix_base_directory(matrixbase);

  if (periodic_bc.provided()) HTM.Periodic_boundary_condition=periodic_bc;
  HTM.select_hamiltonian((std::string)model);

  HTM.get_parameters(par_file.Value);

  if (integration.provided()) ApplyLimit::apply(HTM,integration);
  else HTM.limits(-1.0,30.0,100);
  if (cutoff.provided()) HTM.eigenvalue_cutoff((int)cutoff);

  HTM.temperature(temperature.Value);

  ofstream outfile; outfile.open(((string)output_name+".dat").c_str());
  ofstream verfile; verfile.open(((string)output_name+".ver").c_str());
  ofstream regfile;
  ofstream texfile; texfile.open(((string)output_name+".tex").c_str());

  RegressionModel<value_type> RM(HTM);

  typedef RegressionModel<value_type>::sequence_type sequence_type;
  
  RM.Debug=debug_file;

  RM.prediction_method=prediction_method;
  
  RM.Min_length_regression=min_length_regression;

  if ((string)result=="prediction" && reg_name.provided())
    {
    cout << "Reading regression parameters from " << reg_name.Value << std::endl;
    RM.read_c(reg_name);
    }
  else regfile.open(((string)output_name+".reg").c_str());

  RM.HTM->debug(debug.Value);
  RM.HTM->verbose(debug.Value);

//What should be do if an dataset file is provided
  if(datafile.provided())
    {
    RM.dataset.read_xml((string)datafile,0,outfile);
    if ((string)result == "prediction") RM.predict_all();
    if (exp_err.provided()) RM.dataset.randomise_dataset();
    if((string)result=="nncheck")
      {
      std::cout << RM.Duplex_set << std::endl;
      std::cout << RM.Neighbours_set << std::endl;
      exit(1);
      }
    }
    
  if ((string)result=="averagey")
    {
    if (sequence.provided() && csequence.provided()) 
      {
      HTM.sequence((string)sequence,(string)csequence);
      }   
    else
      {
      if (sequence.provided()) HTM.sequence((string)sequence);
      }
    Looping<ActionAverageY>::result_vector(HTM,0,outfile);
    }

  
  if (sequence.provided() and (string)result!="averagey")
    {
    RegressionModel<>::sequence_info_class si;
    if (!csequence.provided())
      {
      NucleotideSequence<> seq(sequence.Value.length()),comp(sequence.Value.length());
      seq << sequence.Value;
      comp = seq;
      comp.complementary();
      cout << "seq=" << seq << " cseq=" << comp << std::endl;
      si=RM.add_temperature(seq,comp,salt.Value,0.0,0.0);
      }
    else
      {
      si=RM.add_temperature(sequence.Value,csequence.Value,salt.Value,0.0,0.0);
      }
    RM.predict(si,2);
    double temp=0.0, omega=0.0;
    temp=si.temperature.predicted;
    omega=si.thermal_equivalence;
    cout << "omega=" << omega << " Tm=" << temp << std::endl;
    RM.dataset.update(si);
    RM.dataset.print_temperatures(outfile);
    }

  if(datafile.provided() and (string)result!="averagey")
    {
    RM.recalculate_all_thermal_equivalences();
    if ((string)result=="regression") RM.calculate_regression(regfile);
    RM.predict_all();
    RM.dataset.difference();
    RM.verify(verfile);
    RM.dataset.print_temperatures(outfile);
    RM.print_tex(texfile);
    }

  outfile.close();
  std::cout << "Log: " << JobControl::LogFileName << " ";
  JobControl::summary(std::cout) << std::endl;
  JobControl::summary();
  JobControl::status("running",false);

  return 0;
}

